"""
Joint inference module.
"""

__author__ = "Janek Sendrowski"
__contact__ = "sendrowski.janek@gmail.com"
__date__ = "2023-02-26"

import copy
import functools
import logging
import time
from typing import List, Dict, Tuple, Literal, Optional, cast, Callable

import jsonpickle
import multiprocess as mp
import numpy as np
import pandas as pd
from numpy.linalg import norm
from tqdm import tqdm

from .abstract_inference import Inference
from .base_inference import BaseInference
from .bootstrap import Bootstrap
from .config import Config
from .optimization import Optimization, SharedParams, pack_shared, expand_shared, \
    Covariate, flatten_dict, merge_dicts, correct_values, parallelize as parallelize_func, expand_fixed, \
    collapse_fixed, unpack_shared
from .parametrization import Parametrization
from .settings import Settings
from .spectrum import Spectrum, Spectra

# get logger
logger = logging.getLogger('fastdfe')


class JointInference(BaseInference):
    """
    Enabling the sharing of parameters among several Inference objects.

    Example usage:

    ::

        import fastdfe as fd

        # neutral SFS for two types
        sfs_neut = fd.Spectra(dict(
            pendula=[177130, 997, 441, 228, 156, 117, 114, 83, 105, 109, 652],
            pubescens=[172528, 3612, 1359, 790, 584, 427, 325, 234, 166, 76, 31]
        ))

        # selected SFS for two types
        sfs_sel = fd.Spectra(dict(
            pendula=[797939, 1329, 499, 265, 162, 104, 117, 90, 94, 119, 794],
            pubescens=[791106, 5326, 1741, 1005, 756, 546, 416, 294, 177, 104, 41]
        ))

        # create inference object
        inf = fd.JointInference(
            sfs_neut=sfs_neut,
            sfs_sel=sfs_sel,
            fixed_params=dict(eps=0), # fix eps to 0
            do_bootstrap=True
        )

        # run inference
        inf.run()

    """

    def __init__(
            self,
            sfs_neut: Spectra,
            sfs_sel: Spectra,
            include_divergence: bool = None,
            intervals_del: Tuple[float, float, int] = (-1.0e+8, -1.0e-5, 1000),
            intervals_ben: Tuple[float, float, int] = (1.0e-5, 1.0e4, 1000),
            integration_mode: Literal['midpoint', 'quad'] = 'midpoint',
            linearized: bool = True,
            model: Parametrization | str = 'GammaExpParametrization',
            seed: int = 0,
            x0: Dict[str, Dict[str, float]] = {},
            bounds: Dict[str, Tuple[float, float]] = {},
            scales: Dict[str, Literal['lin', 'log', 'symlog']] = {},
            loss_type: Literal['likelihood', 'L2'] = 'likelihood',
            opts_mle: dict = {},
            method_mle: str = 'L-BFGS-B',
            n_runs: int = 10,
            fixed_params: Dict[str, Dict[str, float]] = {},
            shared_params: List[SharedParams] = [],
            covariates: List[Covariate] = [],
            do_bootstrap: bool = False,
            n_bootstraps: int = 100,
            n_bootstrap_retries: int = 2,
            parallelize: bool = True,
            folded: bool = None,
            **kwargs
    ):
        """
        Create instance.

        :param sfs_neut: Neutral SFS. Note that we require monomorphic counts to be specified in order to infer
            the mutation rate.
        :param sfs_sel: Selected SFS. Note that we require monomorphic counts to be specified in order to infer
            the mutation rate.
        :param include_divergence: Whether to include divergence in the likelihood
        :param intervals_del: ``(start, stop, n_interval)`` for deleterious population-scaled
            selection coefficients. The intervals will be log10-spaced.
        :param intervals_ben: Same as ``intervals_del`` but for positive selection coefficients.
        :param integration_mode: Integration mode, ``quad`` not recommended
        :param linearized: Whether to use the linearized model, ``False`` not recommended
        :param model: DFE parametrization
        :param seed: Random seed. Use ``None`` for no seed.
        :param x0: Dictionary of initial values in the form ``{type: {param: value}}``
        :param bounds: Bounds for the optimization in the form {param: (lower, upper)}
        :param scales: Scales for the optimization in the form {param: scale}
        :param loss_type: Loss type
        :param opts_mle: Options for the optimization
        :param method_mle: Method to use for optimization. See `scipy.optimize.minimize` for available methods.
        :param n_runs: Number of independent optimization runs out of which the best one is chosen. The first run
            will use the initial values if specified. Consider increasing this number if the optimization does not
            produce good results.
        :param fixed_params: Dictionary of fixed parameters in the form ``{type: {param: value}}``
        :param shared_params: List of shared parameters
        :param do_bootstrap: Whether to perform bootstrapping
        :param n_bootstraps: Number of bootstraps
        :param n_bootstrap_retries: Number of retries for bootstraps that did not terminate normally.
        :param parallelize: Whether to parallelize computations
        :param folded: Whether the SFS are folded. If not specified, the SFS will be folded if all of the given
            SFS appear to be folded.
        :param kwargs: Additional keyword arguments which are ignored.
        """
        if sfs_neut.has_dots() or sfs_sel.has_dots():
            raise ValueError('Type names cannot contain dots are not allowed as they are used internally.'
                             'You can use the `replace_dots` method to replace dots with another character.')

        # check whether types are equal
        if set(sfs_neut.types) != set(sfs_sel.types):
            raise ValueError('The neutral and selected spectra must have exactly the same types.')

        #: SFS types
        self.types: List[str] = sfs_neut.types

        # initialize parent
        # the `self.folded` property will generalize nicely as we
        # evaluate ``sfs_sel.all.is_folded and sfs_neut.all.is_folded``
        BaseInference.__init__(**locals())

        #: original MLE parameters before adding covariates and unpacked shared
        self.params_mle_raw: Optional[Dict[str, Dict[str, float]]] = None

        #: Shared parameters with expanded 'all' type
        self.shared_params = expand_shared(shared_params, self.types, self.optimization.param_names)

        # add covariates as shared parameters
        self._add_covariates_as_shared(covariates)

        # check if the shared parameters were specified correctly
        self.check_shared_params()

        # throw error if joint inference does not make sense
        if not self.joint_inference_makes_sense():
            self._logger.warning('Joint inference does not make sense as no parameters are shared '
                                 'across more than one type. If this is not intended consider running '
                                 'several marginal inferences instead, which is more efficient.')

        # issue notice
        self._logger.info(f'Using shared parameters {self.shared_params}.')

        #: Covariates indexed by parameter name
        self.covariates: Dict[str, Covariate] = {f"c{i}": cov for i, cov in enumerate(covariates)}

        # only show param and values
        cov_repr = dict((k, dict(param=v.param, values=v.values)) for k, v in self.covariates.items())

        # issue notice
        self._logger.info(f'Including covariates: {cov_repr}.')

        # parameter names for covariates
        args_cov = list(self.covariates.keys())

        # bounds for covariates
        bounds_cov = dict((k, cov.bounds) for k, cov in self.covariates.items())

        #: Initial values for covariates
        self._x0_cov = dict((k, cov.x0) for k, cov in self.covariates.items())

        # use linear scale for covariates
        scales_cov = dict((k, cov.bounds_scale) for k, cov in self.covariates.items())

        #: fixed parameters with expanded 'all' type
        self.fixed_params: Dict[str, Dict[str, float]] = expand_fixed(fixed_params, self.types)

        # collapse fixed parameters
        fixed_collapsed = collapse_fixed(self.fixed_params, self.types)

        # include 'all' type with infers the DFE for all spectra added together
        #: Dictionary of marginal inferences indexed by type
        self.marginal_inferences: Dict[str, BaseInference] = dict(
            all=BaseInference(
                sfs_neut=sfs_neut,
                sfs_sel=sfs_sel,
                discretization=self.discretization,
                include_divergence=include_divergence,
                model=model,
                seed=seed,
                x0=x0,
                bounds=bounds,
                scales=scales,
                loss_type=loss_type,
                opts_mle=opts_mle,
                method_mle=method_mle,
                fixed_params=dict(all=fixed_collapsed['all']) if 'all' in fixed_collapsed else {},
                do_bootstrap=do_bootstrap,
                n_bootstraps=n_bootstraps,
                n_bootstrap_retries=n_bootstrap_retries,
                parallelize=parallelize,
                folded=self.folded,
                n_runs=n_runs)
        )

        # check that the fixed parameters are valid
        self.check_fixed_params_exist()

        # check if the fixed parameters are compatible with the shared parameters
        self.check_no_shared_params_fixed()

        #: parameter scales
        self.scales: Dict[str, Literal['lin', 'log', 'symlog']] = \
            self.model.scales | self._default_scales | scales_cov | scales

        #: parameter bounds
        self.bounds: Dict[str, Tuple[float, float]] = self.model.bounds | self._default_bounds | bounds | bounds_cov

        # create optimization instance for joint inference
        # take initial values and bounds from marginal inferences
        # and from this inference for type 'all'
        #: Joint optimization instance
        self.optimization: Optimization = Optimization(
            bounds=self.bounds,
            scales=self.scales,
            opts_mle=self.optimization.opts_mle,
            method_mle=method_mle,
            loss_type=self.optimization.loss_type,
            param_names=self.model.param_names + ['eps'] + args_cov,
            parallelize=self.parallelize,
            fixed_params=self.fixed_params,
            seed=self.seed
        )

        # Construct the marginal inference object for each type.
        # Note that we use the same discretization
        # instance to avoid precomputing the linearization
        # several times.
        for t in sfs_neut.types:
            self.marginal_inferences[t] = BaseInference(
                # pass subtypes
                sfs_neut=sfs_neut[[t]],
                sfs_sel=sfs_sel[[t]],
                discretization=self.discretization,
                include_divergence=include_divergence,
                model=model,
                seed=seed,
                x0=x0,
                bounds=bounds,
                scales=scales,
                loss_type=loss_type,
                opts_mle=opts_mle,
                method_mle=method_mle,
                fixed_params=dict(all=self.fixed_params[t]) if t in self.fixed_params else {},
                do_bootstrap=do_bootstrap,
                n_bootstraps=n_bootstraps,
                n_bootstrap_retries=n_bootstrap_retries,
                parallelize=parallelize,
                folded=self.folded,
                n_runs=n_runs
            )

        #: Joint inference object indexed by type
        self.joint_inferences: Dict[str, BaseInference] = {}

        # add base inference object for each type
        for t in sfs_neut.types:
            self.joint_inferences[t] = BaseInference(
                # pass subtypes
                sfs_neut=sfs_neut[[t]],
                sfs_sel=sfs_sel[[t]],
                discretization=self.discretization,
                include_divergence=include_divergence,
                model=model,
                seed=seed,
                x0=x0,
                bounds=bounds,
                scales=scales,
                loss_type=loss_type,
                opts_mle=opts_mle,
                method_mle=method_mle,
                fixed_params=dict(all=self.fixed_params[t]) if t in self.fixed_params else {},
                do_bootstrap=do_bootstrap,
                n_bootstraps=n_bootstraps,
                n_bootstrap_retries=n_bootstrap_retries,
                parallelize=parallelize,
                folded=self.folded,
                n_runs=n_runs,
                locked=True
            )

    def check_shared_params(self):
        """
        Check if the shared parameters were specified correctly.
        """
        for shared in self.shared_params:
            # check parameters
            if len(set(shared.params) - set(self.param_names)) != 0:
                raise ValueError(
                    f'Specified shared parameters {list(set(shared.params) - set(self.param_names))} '
                    f"don't match with any of {self.param_names}."
                )

            # check types
            if len(set(shared.types) - set(self.types)) != 0:
                raise ValueError(
                    f'Specified types {list(set(shared.types) - set(self.types))} '
                    f" in shared parameters don't match with any of {self.types}."
                )

    def check_no_shared_params_fixed(self):
        """
        Check that no shared parameters are fixed and raise an error otherwise.
        """
        for shared in self.shared_params:
            for t in shared.types:
                if t in self.fixed_params:
                    for p in shared.params:
                        if p in self.fixed_params[t]:
                            raise ValueError(f"Parameter '{p}' in type '{t}' is both "
                                             f"shared and fixed, which is not allowed. "
                                             f"Note that covariates are automatically shared.")

    def get_shared_param_names(self) -> List[str]:
        """
        Get the names of the shared parameters.

        :return: Names of the shared parameters.
        """
        shared = []

        for p in self.shared_params:
            shared += p.params

        return cast(List[str], np.unique(shared).tolist())

    def _add_covariates_as_shared(self, covariates: List[Covariate]):
        """
        Add covariates as shared parameters.

        :param covariates: List of covariates.
        """
        cov_but_not_shared = self._determine_unshared_covariates(covariates)

        # add parameters with covariates to shared parameters
        if len(cov_but_not_shared) > 0:
            self._logger.info(f'Parameters {cov_but_not_shared} have '
                              f'covariates and thus need to be shared. '
                              f'Adding them to shared parameters.')

            # add to shared parameters
            self.shared_params.append(SharedParams(params=cov_but_not_shared, types=self.types))

    def _determine_unshared_covariates(self, covariates):
        """
        Determine which covariates are not shared.

        :param covariates:
        :return:
        """
        # determine completely shared parameters
        completely_shared = []
        for shared in self.shared_params:
            if len(shared.types) == len(self.types):
                completely_shared += shared.params

        # determine parameters with covariates
        params_with_covariates = [cov.param for cov in covariates]

        # determine parameters with covariates that are not shared
        return list(set(params_with_covariates) - set(completely_shared))

    def run(
            self,
            **kwargs
    ) -> Spectrum:
        """
        Run inference.

        :param kwargs: Additional keyword arguments
        :return: Modelled SFS.
        """

        # run marginal optimization
        self._run_marginal()

        # run joint optimization
        return self._run_joint()

    def _run_marginal(self):
        """
        Run marginal optimization.

        :return: Dict of marginal inferences indexed by type.
        """

        def run_marginal(data: Tuple[str, BaseInference]) -> Tuple[str, BaseInference]:
            """
            Run marginal inference. Start with initial values
            obtained from 'all' type.

            :param data: Tuple of type and inference object.
            :return: Tuple of type and inference object.
            """
            # issue notice
            self._logger.info(f"Running marginal inference for type '{data[0]}'.")

            data[1].run_if_required(
                do_bootstrap=False,
                pbar=False
            )

            return data

        # run for 'all' type
        run_marginal(('all', self.marginal_inferences['all']))

        # update initial values of marginal inferences
        # with MLE estimate of 'all' type
        for inf in self.marginal_inferences.values():
            inf.x0 = dict(all=self.marginal_inferences['all'].params_mle)

        # skip marginal inference if only one type was specified
        if len(self.types) == 1:
            self._logger.info('Skipping marginal inference as only one type was specified.')
        else:
            # issue notice
            self._logger.info(f'Running marginal inferences for types {self.types}.')

        # optionally parallelize marginal inferences
        run_inferences = dict(parallelize_func(
            func=run_marginal,
            data=list(self.marginals_without_all().items()),
            parallelize=False,
            pbar=False
        ))

        # reassign marginal inferences
        self.marginal_inferences = dict(all=self.marginal_inferences['all']) | run_inferences

        return self.marginal_inferences

    def marginals_without_all(self) -> Dict[str, BaseInference]:
        """
        Get marginal inference without 'all' type.

        :return: Dict of marginal inferences indexed by type.
        """
        return dict((t, inf) for t, inf in self.marginal_inferences.items() if t != 'all')

    def _run_joint(self) -> Spectrum:
        """
        Run joint optimization.

        :return: Modelled SFS.
        """
        # issue notice
        self._logger.info(f'Running joint inference for types {self.types}.')

        # issue notice
        self._logger.debug(f'Starting numerical optimization of {self.n_runs} '
                           'independently initialized samples which are run ' +
                           ('in parallel.' if self.parallelize else 'sequentially.'))

        # starting time of joint inference
        start_time = time.time()

        # Perform joint optimization.
        self.result, params_mle = self.optimization.run(
            x0=self.get_x0(),
            scales=self.scales,
            bounds=self.bounds,
            n_runs=self.n_runs,
            get_counts=self.get_counts(),
            desc=f'{self.__class__.__name__}>Performing joint inference'
        )

        # assign likelihoods
        self.likelihoods = self.optimization.likelihoods

        # store packed MLE params for later usage
        self.params_mle_raw = copy.deepcopy(params_mle)

        # unpack shared parameters
        params_mle = unpack_shared(params_mle)

        # normalize parameters for each type
        for t in self.types:
            params_mle[t] = self.model._normalize(params_mle[t])

        # report on optimization result
        self._report_result(self.result, params_mle)

        # assign optimization result and MLE parameters for each type
        for t, inf in self.joint_inferences.items():
            params_mle[t] = correct_values(
                params=Covariate._apply(self.covariates, params_mle[t], t),
                bounds=self.bounds,
                scales=self.scales
            )

            # remove effect of covariates and assign result
            inf._assign_result(self.result, params_mle[t])

            # assign execution time
            inf.execution_time = time.time() - start_time

        # assign MLE params
        self.params_mle = params_mle

        # assign joint likelihood
        self.likelihood = -self.result.fun

        # calculate L2 residual
        self.L2_residual = self.get_residual(2)

        # check L2 residual
        self._check_L2_residual()

        # add execution time
        self.execution_time += time.time() - start_time

        # perform bootstrap if configured
        if self.do_bootstrap:
            self.bootstrap()

        return self.sfs_mle

    def joint_inference_makes_sense(self) -> bool:
        """
        Check if joint inference makes sense.

        :return: Whether joint inference makes sense.
        """
        return len(self.types) >= 2 and len(self.shared_params) != 0

    def get_counts(self) -> dict:
        """
        Get callback functions for modelling SFS counts from given parameters.

        :return: Dict of callback functions indexed by type.
        """
        # Note that it's important we bind t into the lambda function
        # at the time of creation.
        return dict((t, (lambda params, t=t: inf._model_sfs(
            discretization=self.discretization,
            model=self.model,
            params=correct_values(Covariate._apply(self.covariates, params, t), self.bounds, self.scales),
            sfs_neut=self.joint_inferences[t].sfs_neut,
            sfs_sel=self.joint_inferences[t].sfs_sel,
            folded=self.folded
        ))) for t, inf in self.joint_inferences.items())

    @BaseInference._run_if_required_wrapper
    @functools.lru_cache
    def run_joint_without_covariates(self, do_bootstrap: bool = True) -> 'JointInference':
        """
        Run joint inference without covariates. Note that the result of this function is cached.

        :return: Joint inference instance devoid of covariates.
        """
        config = self.create_config()

        # retain shared parameters but remove covariates
        config.update(
            shared_params=self.shared_params,
            covariates={},
            do_bootstrap=do_bootstrap
        )

        # create copy
        other = JointInference.from_config(config)

        # issue notice
        self._logger.info('Running joint inference without covariates.')

        # run inference
        other.run()

        return other

    def create_config(self) -> 'Config':
        """
        Create a config object from the inference object.

        :return: Config object.
        """
        return Config(
            sfs_neut=Spectra.from_spectra(dict((t, inf.sfs_neut) for t, inf in self.marginals_without_all().items())),
            sfs_sel=Spectra.from_spectra(dict((t, inf.sfs_sel) for t, inf in self.marginals_without_all().items())),
            intervals_ben=self.discretization.intervals_ben,
            intervals_del=self.discretization.intervals_del,
            integration_mode=self.discretization.integration_mode,
            linearized=self.discretization.linearized,
            model=self.model,
            seed=self.seed,
            opts_mle=self.optimization.opts_mle,
            method_mle=self.optimization.method_mle,
            x0=self.x0,
            bounds=self.bounds,
            scales=self.scales,
            loss_type=self.optimization.loss_type,
            fixed_params=self.fixed_params,
            covariates=[c for c in self.covariates.values()],
            shared_params=self.shared_params,
            do_bootstrap=self.do_bootstrap,
            n_bootstraps=self.n_bootstraps,
            parallelize=self.parallelize,
            n_runs=self.n_runs
        )

    @BaseInference._run_if_required_wrapper
    def perform_lrt_covariates(self, do_bootstrap: bool = True) -> float:
        """
        Perform likelihood ratio test against joint inference without covariates.
        In the simple model we share parameters across types. Low p-values indicate that
        including covariates significantly improves the model fit.

        To access the JointInference object without covariates, you can call :meth:`run_joint_without_covariates`,
        which is cached.

        :param do_bootstrap: Whether to bootstrap. This improves the accuracy of the p-value. Note
            that if bootstrapping was performed previously without updating the likelihood, this won't have any effect.
        :return: Likelihood ratio test p-value.
        """
        if len(self.covariates) == 0:
            raise ValueError('No covariates were specified.')

        # bootstrap if required
        if do_bootstrap:
            self.bootstrap_if_required()

        # run joint inference without covariates
        simple = self.run_joint_without_covariates(do_bootstrap=do_bootstrap)

        return self.lrt(simple.likelihood, self.likelihood, len(self.covariates))

    def _get_run_bootstrap_sample(self) -> Callable[[int], Tuple['scipy.optimize.OptimizeResult', dict]]:
        """
        Get function which runs a single bootstrap sample.

        :return: Static function which runs a single bootstrap sample, taking an optional seed and returning the
            optimization result and the MLE parameters.
        """
        optimization = self.optimization
        discretization = self.discretization
        covariates = self.covariates
        model = self.model
        types = self.types
        folded = self.folded
        params_mle_raw = self.params_mle_raw
        scales_linear = self.get_scales_linear()
        bounds_linear = self.get_bounds_linear()
        scales = self.scales
        bounds = self.bounds
        n_retries = self.n_bootstrap_retries
        sfs_neut = dict((t, self.marginal_inferences[t].sfs_neut) for t in self.types)
        sfs_sel = dict((t, self.marginal_inferences[t].sfs_sel) for t in self.types)

        def run_bootstrap_sample(seed: int) -> ('scipy.optimize.OptimizeResult', dict):
            """
            Resample the observed selected SFS and rerun the optimization procedure.
            We take the MLE params as initial params here.
            We make this function static to improve performance when parallelizing.
            In case the optimization does not terminate normally, we retry up to `n_retries` times.

            :return: Optimization result and dictionary of MLE params
            """
            result, params_mle = None, None

            # retry up to `n_retries` times
            for i in range(max(n_retries, 0) + 1):

                # resample spectra
                spectra = {}
                for t in types:
                    spectra[t] = dict(
                        neut=sfs_neut[t].resample(seed=seed + 2 * i),
                        sel=sfs_sel[t].resample(seed=seed + 2 * i + 1)
                    )

                # perform joint optimization
                # Note that it's important we bind t into the lambda function
                # at the time of creation.
                result, params_mle = optimization.run(
                    x0=params_mle_raw,
                    scales=scales_linear,
                    bounds=bounds_linear,
                    n_runs=1,
                    debug_iterations=False,
                    print_info=False,
                    desc=f"{self.__class__.__name__}>Bootstrapping joint inference",
                    get_counts=dict((t, lambda params, t=t: BaseInference._model_sfs(
                        discretization=discretization,
                        model=model,
                        params=correct_values(Covariate._apply(covariates, params, t), bounds, scales),
                        sfs_neut=spectra[t]['neut'],
                        sfs_sel=spectra[t]['sel'],
                        folded=folded
                    )) for t in types)
                )

                # unpack shared parameters
                params_mle = unpack_shared(params_mle)

                for t in types:
                    # normalize parameters for each type
                    params_mle[t] = model._normalize(params_mle[t])

                    # add covariates for each type
                    params_mle[t] = correct_values(
                        params=Covariate._apply(covariates, params_mle[t], t),
                        bounds=bounds,
                        scales=scales
                    )

                if result.success:
                    return result, params_mle

            return result, params_mle

        return run_bootstrap_sample

    def bootstrap(
            self,
            n_samples: int = None,
            parallelize: bool = None,
            n_retries: int = None,
            update_likelihood: bool = True,
            **kwargs
    ) -> Optional[pd.DataFrame]:
        """
        Perform the parametric bootstrap both for the marginal and joint inferences.

        :param n_samples: Number of bootstrap samples. Defaults to :attr:`n_bootstraps`.
        :param parallelize: Whether to parallelize computations. Defaults to :attr:`parallelize`.
        :param n_retries: Number of retries for bootstraps that did not terminate normally. Defaults to
            :attr:`n_bootstrap_retries`.
        :param update_likelihood: Whether to update the likelihood
        :return: DataFrame with bootstrap samples
        """
        # perform inference first if not done yet
        self.run_if_required()

        # update properties
        self.update_properties(
            n_bootstraps=n_samples,
            parallelize=parallelize,
            n_bootstrap_retries=n_retries
        )

        n_bootstraps = int(self.n_bootstraps)

        with tqdm(
                total=len(self.marginal_inferences) * n_bootstraps,
                disable=Settings.disable_pbar,
                desc=f"{self.__class__.__name__}>Bootstrapping marginal inferences"
        ) as pbar:

            # bootstrap marginal inferences
            for t, inf in self.marginal_inferences.items():
                self._logger.info(f"Bootstrapping type '{t}'.")

                inf.bootstrap(
                    n_samples=n_bootstraps,
                    parallelize=self.parallelize,
                    n_retries=self.n_bootstrap_retries,
                    update_likelihood=update_likelihood,
                    pbar=False
                )

                pbar.update(n_bootstraps)

        start_time = time.time()

        # parallelize computations if desired
        if self.parallelize:

            self._logger.debug(f"Running {n_bootstraps} joint bootstrap samples "
                               f"in parallel on {min(mp.cpu_count(), n_bootstraps)} cores.")

        else:
            self._logger.debug(f"Running {n_bootstraps} joint bootstrap samples sequentially.")

        # seeds for bootstraps
        seeds = self.rng.integers(0, high=2 ** 32, size=n_bootstraps)

        # run bootstraps
        result = parallelize_func(
            func=self._get_run_bootstrap_sample(),
            data=seeds,
            parallelize=self.parallelize,
            pbar=True,
            desc=f"{self.__class__.__name__}>Bootstrapping joint inference"
        )

        # number of successful runs
        n_success = np.sum([res.success for res in result[:, 0]])

        # issue warning if some runs did not finish successfully
        if n_success < n_bootstraps:
            self._logger.warning(
                f"{n_bootstraps - n_success} out of {n_bootstraps} bootstrap samples "
                "did not terminate normally during numerical optimization. "
                "The confidence intervals might thus be unreliable. Consider "
                "increasing the number of retries (`n_retries`), "
                "adjusting the optimization parameters (increasing `gtol` or `n_runs`), "
                "or decreasing the number of optimized parameters."
            )

        # dataframe of MLE estimates in flattened format
        self.bootstraps = pd.DataFrame([flatten_dict(r) for r in result[:, 1]])

        # assign bootstrap results
        self.bootstrap_results = list(result[:, 0])

        # assign bootstrap parameters to joint inference objects
        for t, inf in self.joint_inferences.items():
            # filter for columns belonging to the current type
            inf.bootstraps = self.bootstraps.filter(regex=f'{t}\\..*').rename(columns=lambda x: x.split('.')[-1])

            # add estimates for alpha to the bootstraps
            inf._add_alpha_to_bootstraps()

        # add execution time
        self.execution_time += time.time() - start_time

        # assign average likelihood of successful runs
        if update_likelihood:
            self.likelihood = np.mean([-res.fun for res in result[:, 0] if res.success] + [self.likelihood])

        return self.bootstraps

    def bootstrap_if_required(self):
        """
        Bootstrap if not done yet.
        """
        if self.bootstraps is None:
            self.bootstrap()

    def get_x0(self) -> Dict[str, Dict[str, float]]:
        """
        Get initial values for joint inference.

        :return: Dictionary of initial values indexed by type
        """
        x0 = {}

        # create initial values from marginal inferences
        for t, inf in self.marginals_without_all().items():

            # the MLE params might not be defined
            if inf.params_mle is not None:
                x0[t] = inf.params_mle
            else:
                x0[t] = self.marginal_inferences['all'].x0['all']

        # get shared parameters from last inference and merge
        # with parameters for type 'all'
        shared = {}

        # get dict of shared parameters
        for s in self.shared_params:
            for p in s.params:
                # take mean value over types
                shared[':'.join(s.types) + '.' + p] = cast(float, np.mean([x0[t][p] for t in s.types]))

        # pack shared parameters
        packed = pack_shared(x0, self.shared_params, shared)

        # add parameters for covariates and return
        return merge_dicts(packed, {':'.join(self.types): self._x0_cov})

    @BaseInference._run_if_required_wrapper
    def perform_lrt_shared(self, do_bootstrap: bool = True) -> float:
        """
        Compare likelihood of joint inference with product of marginal likelihoods.
        This provides information about the goodness of fit achieved by the parameter sharing.
        Low p-values indicate that parameter sharing is not justified, i.e., that the marginal
        inferences provide a significantly better fit to the data. Note that it is more difficult to properly
        optimize the joint likelihood, which makes this test conservative, i.e., the reported p-value
        may be larger than what it should be.

        :param do_bootstrap: Whether to perform bootstrapping. This improves the accuracy of the p-value. Note
            that if bootstrapping was performed previously without updating the likelihood, this won't have any effect.
        :return: p-value
        """
        if do_bootstrap:
            self.bootstrap_if_required()

        # determine likelihood of marginal inferences
        ll_marginal = sum([inf.likelihood for inf in self.marginals_without_all().values()])

        # determine number of parameters
        n_marginal = np.sum([len(inf.params_mle) for inf in self.marginals_without_all().values()])
        n_joint = len(flatten_dict(self.get_x0()))

        return self.lrt(ll_simple=self.likelihood, ll_complex=ll_marginal, df=n_marginal - n_joint)

    def get_inferences(
            self,
            types: List[str] = None,
            labels: List[str] = None,
            show_marginals: bool = True
    ) -> Dict[str, 'BaseInference']:
        """
        Get all inference objects as dictionary.

        :param types: Types to include
        :param labels: Labels for types
        :param show_marginals: Whether to also show marginal inferences
        :return: Dictionary of base inference objects indexed by type and joint vs marginal subtypes
        """

        def get(infs: Dict[str, BaseInference], prefix: str) -> Dict[str, BaseInference]:
            """
            Get filtered and prefixed inferences.

            :param infs: Dictionary of inferences.
            :param prefix: Prefix
            :return: Dictionary of inferences
            """
            # filter types if types are given
            if types is not None:
                infs = dict((k, v) for k, v in infs.items() if k in types)

            # add prefix to keys
            return dict((prefix + '.' + k, v) for k, v in infs.items())

        # get joint inferences
        inferences = get(self.joint_inferences, 'joint')

        if show_marginals:
            # include marginal inferences
            inferences = get(self.marginal_inferences, 'marginal') | inferences

        # use labels as keys if given
        if labels is not None:

            if len(labels) != len(inferences):
                raise ValueError(f'Number of labels ({len(labels)}) does not match '
                                 f'number of inferences ({len(inferences)}).')

            inferences = dict(zip(labels, inferences.values()))

        return inferences

    @BaseInference._run_if_required_wrapper
    def plot_discretized(
            self,
            file: str = None,
            show: bool = True,
            intervals: np.ndarray = np.array([-np.inf, -100, -10, -1, 0, 1, np.inf]),
            show_marginals: bool = True,
            confidence_intervals: bool = True,
            ci_level: float = 0.05,
            bootstrap_type: Literal['percentile', 'bca'] = 'percentile',
            title: str = 'discretized DFE comparison',
            labels: List[str] = None,
            kwargs_legend: dict = dict(prop=dict(size=8)),
            ax: 'plt.Axes' = None
    ) -> 'plt.Axes':
        """
        Plot discretized DFE comparing the different types.

        :param labels: Labels for types
        :param title: Title of plot
        :param intervals: Array of interval boundaries over ``(-inf, inf)`` yielding ``intervals.shape[0] - 1`` bars.
        :param show_marginals: Whether to also show marginal inferences
        :param bootstrap_type: Type of bootstrap
        :param ci_level: Confidence level
        :param confidence_intervals: Whether to plot confidence intervals
        :param file: File to save plot to
        :param show: Whether to show plot
        :param ax: Axes to plot on. Only for Python visualization backend.
        :param kwargs_legend: Keyword arguments passed to :meth:`plt.legend`. Only for Python visualization backend.
        :return: Axes object
        """
        labels, inferences = zip(*self.get_inferences(labels=labels, show_marginals=show_marginals).items())

        return Inference.plot_discretized(**locals())

    def plot_sfs_comparison(
            self,
            sfs_types: List[Literal['modelled', 'observed', 'selected', 'neutral']] = ['observed', 'modelled'],
            types: List[str] = None,
            labels: List[str] = None,
            colors: List[str] = None,
            file: str = None,
            show: bool = True,
            ax: 'plt.Axes' = None,
            title: str = 'SFS comparison',
            use_subplots: bool = False,
            show_monomorphic: bool = False,
            kwargs_legend: dict = dict(prop=dict(size=8)),

    ) -> 'plt.Axes':
        """
        Plot SFS comparison.

        :param types: Types to plot
        :param file: File to save plot to
        :param labels: Labels for types.
        :param colors: Colors for types. Only for Python visualization backend.
        :param sfs_types: Types of SFS to plot
        :param show: Whether to show plot
        :param ax: Axes to plot on. Only for Python visualization backend and if ``use_subplots`` is ``False``.
        :param title: Plot title
        :param use_subplots: Whether to use subplots. Only for Python visualization backend.
        :param show_monomorphic: Whether to show monomorphic counts
        :param kwargs_legend: Keyword arguments passed to :meth:`plt.legend`. Only for Python visualization backend.
        :return: Axes object
        """
        from .visualization import Visualization

        if 'modelled' in sfs_types:
            self.run_if_required()

        mapping = dict(
            observed='sfs_sel',
            selected='sfs_sel',
            modelled='sfs_mle',
            neutral='sfs_neut'
        )

        inferences = self.get_inferences(types=types)

        def get_label(t: str, sfs_type: str) -> str:
            """
            Get label for types.

            :param t: Type
            :param sfs_type: SFS type
            :return: Label
            """
            subtypes = t.split('.')

            # insert sfs type at second position
            return '.'.join([subtypes[0]] + [sfs_type] + subtypes[1:])

        # get spectra
        spectra = {get_label(t, sfs): getattr(inf, mapping[sfs]) for t, inf in inferences.items() for sfs in sfs_types}

        return Visualization.plot_spectra(
            spectra=[list(v) for v in spectra.values()],
            labels=list(spectra.keys()) if labels is None else labels,
            colors=colors,
            file=file,
            show=show,
            ax=ax,
            title=title,
            use_subplots=use_subplots,
            show_monomorphic=show_monomorphic,
            kwargs_legend=kwargs_legend
        )

    @BaseInference._run_if_required_wrapper
    def plot_continuous(
            self,
            file: str = None,
            show: bool = True,
            intervals: np.ndarray = None,
            confidence_intervals: bool = True,
            ci_level: float = 0.05,
            bootstrap_type: Literal['percentile', 'bca'] = 'percentile',
            title: str = 'DFE comparison',
            labels: List[str] = None,
            scale_density: bool = False,
            scale: Literal['lin', 'log', 'symlog'] = 'lin',
            kwargs_legend: dict = dict(prop=dict(size=8)),
            ax: 'plt.Axes' = None
    ) -> 'plt.Axes':
        """
        Plot discretized DFE. The special constants ``np.inf`` and ``-np.inf`` are also valid interval bounds.
        By default, the PDF is plotted as is. Due to the logarithmic scale on
        the x-axis, we may get a wrong intuition on how the mass is distributed,
        however. To get a better intuition, we can optionally scale the density
        by the x-axis interval size using ``scale_density = True``. This has the
        disadvantage that the density now changes for x, so that even a constant
        density will look warped.

        :param scale_density: Whether to scale the density by the x-axis interval size
        :param scale: y-scale
        :param labels: Labels for types
        :param title: Title of plot
        :param bootstrap_type: Type of bootstrap
        :param ci_level: Confidence level
        :param confidence_intervals: Whether to plot confidence intervals
        :param file: File to save plot to
        :param show: Whether to show plot
        :param intervals: Array of interval boundaries yielding ``intervals.shape[0] - 1`` bins.
        :param kwargs_legend: Keyword arguments passed to :meth:`plt.legend`. Only for Python visualization backend.
        :param ax: Axes to plot on. Only for Python visualization backend.
        :return: Axes object
        """
        if intervals is None:
            intervals = self.discretization.bins

        labels, inferences = zip(*self.get_inferences(labels=labels).items())

        return Inference.plot_continuous(**locals())

    @BaseInference._run_if_required_wrapper
    def plot_inferred_parameters(
            self,
            file: str = None,
            confidence_intervals: bool = True,
            bootstrap_type: Literal['percentile', 'bca'] = 'percentile',
            ci_level: float = 0.05,
            show: bool = True,
            title: str = 'inferred parameters',
            labels: List[str] = None,
            ax: 'plt.Axes' = None,
            scale: Literal['lin', 'log', 'symlog'] = 'log',
            kwargs_legend: dict = dict(prop=dict(size=8), loc='upper right'),
            **kwargs: List[str]
    ) -> 'plt.Axes':
        """
        Plot discretized DFE comparing the different types.

        :param labels: Labels for types
        :param title: Title of plot
        :param bootstrap_type: Type of bootstrap
        :param ci_level: Confidence level
        :param confidence_intervals: Whether to plot confidence intervals
        :param file: File to save plot to
        :param show: Whether to show plot
        :param ax: Axes to plot on. Only for Python visualization backend.
        :param scale: y-scale
        :param kwargs_legend: Keyword arguments passed to :meth:`plt.legend`. Only for Python visualization backend.
        :return: Axes object
        """
        labels, inferences = zip(*self.get_inferences(labels=labels).items())

        return Inference.plot_inferred_parameters(**locals())

    @BaseInference._run_if_required_wrapper
    def plot_inferred_parameters_boxplot(
            self,
            file: str = None,
            show: bool = True,
            title: str = 'inferred parameters',
            labels: List[str] = None,
            **kwargs: List[str]
    ) -> 'plt.Axes':
        """
        Plot discretized DFE comparing the different types.

        :param labels: Labels for types
        :param title: Title of plot
        :param file: File to save plot to
        :param show: Whether to show plot
        :return: Axes object
        :raises ValueError: If no inference objects are given or no bootstraps are found.
        """
        labels, inferences = zip(*self.get_inferences(labels=labels).items())

        return Inference.plot_inferred_parameters_boxplot(**locals())

    @BaseInference._run_if_required_wrapper
    def plot_covariate(
            self,
            index: int = 0,
            file: str = None,
            show: bool = True,
            title: str = None,
            bootstrap_type: Literal['percentile', 'bca'] = 'percentile',
            show_types: bool = True,
            ci_level: float = 0.05,
            xlabel: str = "cov",
            ylabel: str = None,
            ax: 'plt.Axes' = None
    ) -> 'plt.Axes':
        """
        Plot the covariate given by the index.

        :param index: The index of the covariate.
        :param file: File to save plot to.
        :param show: Whether to show plot.
        :param title: Plot title.
        :param bootstrap_type: Bootstrap type.
        :param show_types: Whether to show types on second x-axis.
        :param ci_level: Confidence level.
        :param xlabel: X-axis label.
        :param ylabel: Y-axis label, defaults to the covariate parameter name.
        :param ax: Axes to plot on. Only for Python visualization backend.
        :return: Axes object.
        """
        from .visualization import Visualization

        key = f"c{index}"

        # check if covariate exists
        if key not in self.covariates:
            raise ValueError(f"Covariate with index {index} does not exist.")

        # default title
        if title is None:
            title = f'covariate {key}'

        cov = self.covariates[key]
        values = [self.params_mle[t][cov.param] for t in self.types]

        # get errors if bootstrapped
        errors = None
        if self.bootstraps is not None:
            bootstraps = np.array([self.bootstraps[f"{t}.{cov.param}"] for t in self.types]).T

            # compute errors
            errors = Bootstrap.get_errors(
                values=values,
                bs=bootstraps,
                bootstrap_type=bootstrap_type,
                ci_level=ci_level
            )[0]

            # take mean of bootstraps as values
            values = bootstraps.mean(axis=0)

        return Visualization.plot_covariate(
            covariates=[cov.values[t] for t in self.types],
            values=values,
            errors=errors,
            file=file,
            show=show,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel or cov.param,
            labels=self.types if show_types else None,
            ax=ax
        )

    def get_cis_params_mle(
            self,
            bootstrap_type: Literal['percentile', 'bca'] = 'percentile',
            ci_level: float = 0.05,
            param_names: Optional[list[str]] = None
    ):
        """
        Get confidence intervals for the parameters.

        :return: Confidence intervals for the parameters
        """
        # get dict of inferences
        inferences = self.get_inferences()

        # get param names if not given
        if param_names is None:
            param_names = list(inferences.values())[0].get_bootstrap_param_names()

        return Inference.get_cis_params_mle(
            inferences=list(inferences.values()),
            bootstrap_type=bootstrap_type,
            ci_level=ci_level,
            param_names=param_names,
            labels=list(inferences.keys())
        )

    def get_discretized(
            self,
            intervals: np.ndarray = np.array([-np.inf, -100, -10, -1, 0, 1, np.inf]),
            confidence_intervals: bool = True,
            ci_level: float = 0.05,
            bootstrap_type: Literal['percentile', 'bca'] = 'percentile'
    ) -> Dict[str, Tuple[np.ndarray, np.ndarray]]:
        """
        Get discretized DFEs.

        :param bootstrap_type: Type of bootstrap
        :param ci_level: Confidence interval level
        :param confidence_intervals: Whether to return confidence intervals
        :param intervals: Array of interval boundaries over ``(-inf, inf)`` yielding ``intervals.shape[0] - 1`` bins.
        :return: Dictionary of array of values and array of errors indexed by inference type
        """
        return Inference.get_discretized(
            inferences=list(self.get_inferences().values()),
            labels=list(self.get_inferences().keys()),
            intervals=intervals,
            confidence_intervals=confidence_intervals,
            ci_level=ci_level,
            bootstrap_type=bootstrap_type
        )

    def get_bootstrap_params(self) -> Dict[str, float]:
        """
        Get bootstrap parameters.

        :return: Bootstrap parameters
        """
        return flatten_dict(dict((t, self.joint_inferences[t].get_bootstrap_params()) for t in self.types))

    def to_json(self) -> str:
        """
        Serialize object. Note that the deserialized inference objects no
        longer share the same optimization instance among other things.

        :return: JSON string
        """
        # using make_ref=True resulted in weird behaviour when unserializing.
        return jsonpickle.encode(self, indent=4, warn=True, make_refs=False)

    @functools.cached_property
    def alpha(self) -> Optional[float]:
        """
        The is no single alpha for the joint inference. Please refer
        to the ``self.joint_inferences[t].alpha``.

        :return: None
        """
        return

    def _set_fixed_params(self, params: Dict[str, Dict[str, float]]):
        """
        Set fixed parameters.

        :param params: Fixed parameters
        """
        # set for 'all' type
        self.marginal_inferences['all']._set_fixed_params(params)

        # expand types
        self.fixed_params = expand_fixed(params, self.types)

        # propagate to inference objects
        for t in self.types:
            self.marginal_inferences[t]._set_fixed_params(dict(all=self.fixed_params[t]))
            self.joint_inferences[t]._set_fixed_params(dict(all=self.fixed_params[t]))

        # propagate to optimization
        self.optimization.set_fixed_params(self.fixed_params)

        # check if the fixed parameters are compatible with the shared parameters
        self.check_no_shared_params_fixed()

    def get_residual(self, k: int) -> float:
        """
        Residual of joint inference. We calculate the residual over the jointly inferred SFS for all types.

        :param k: Order of the norm
        :return: L2 residual
        """
        counts_mle = np.array([inf.sfs_mle.polymorphic for inf in self.joint_inferences.values()]).flatten()
        counts_sel = np.array([inf.sfs_sel.polymorphic for inf in self.joint_inferences.values()]).flatten()

        return norm(counts_mle - counts_sel, k)



"""
JSON handlers.
"""

__author__ = "Janek Sendrowski"
__contact__ = "sendrowski.janek@gmail.com"
__date__ = "2023-02-26"

import json
import logging

import numpy as np
import pandas as pd
from jsonpickle.handlers import BaseHandler

from .optimization import SharedParams, Covariate
from .parametrization import Parametrization
from .spectrum import Spectrum, Spectra

# configure logger
logger = logging.getLogger('fastdfe')


class NumpyArrayHandler(BaseHandler):
    """
    Handler for numpy arrays.
    """

    def flatten(self, x: np.ndarray, data: dict) -> dict:
        """
        Convert Spectrum to dict.

        :param x: Numpy array
        :param data: Dictionary
        :return: Simplified dictionary
        """
        return data | dict(data=x.tolist())

    def restore(self, data: dict) -> np.ndarray:
        """
        Restore Spectrum.

        :param data: Dictionary
        :return: Numpy array
        """
        return np.array(data['data'])


class SpectrumHandler(BaseHandler):
    """
    Handler for spectrum objects.
    """

    def flatten(self, sfs: Spectrum, data: dict) -> dict:
        """
        Convert Spectrum to dict.

        :param sfs: Spectrum object
        :param data: Dictionary
        :return: Simplified dictionary
        """
        return data | dict(data=sfs.to_list())

    def restore(self, data: dict) -> Spectrum:
        """
        Restore Spectrum.

        :param data: Dictionary
        :return: Spectrum object
        """
        return Spectrum.from_list(data['data'])


class SpectraHandler(BaseHandler):
    """
    Handler for spectra objects.
    """

    def flatten(self, sfs: Spectra, data: dict) -> dict:
        """
        Convert Spectra to dict.

        :param sfs: Spectra object
        :param data: Dictionary
        :return: Simplified dictionary
        """
        return data | dict(data=sfs.to_dict())

    def restore(self, data: dict) -> Spectra:
        """
        Restore Spectra.

        :param data: Dictionary
        :return: Spectra object
        """
        return Spectra.from_dict(data['data'])


class DataframeHandler(BaseHandler):
    """
    There were also problems with dataframes, hence the custom handler.
    """

    def flatten(self, df: pd.DataFrame, data: dict) -> dict:
        """
        Convert dataframe to dict.

        :param df: Dataframe
        :param data: Dictionary
        :return: Simplified dictionary
        """
        return data | dict(data=df.to_dict())

    def restore(self, data: dict) -> pd.DataFrame:
        """
        Restore dataframe.

        :param data: Dictionary
        :return: Dataframe
        """
        return pd.DataFrame(data['data'])


class CustomEncoder(json.JSONEncoder):
    """
    Convert numpy arrays and objects to lists and primitives.
    """

    def default(self, obj):
        """
        Convert numpy arrays and objects to lists and primitives.

        :param obj: Object
        :return: Simplified object
        """
        from scipy.optimize._lbfgsb_py import LbfgsInvHessProduct

        if isinstance(obj, Spectrum):
            return obj.to_list()

        if isinstance(obj, Spectra):
            return obj.to_dict()

        if isinstance(obj, np.ndarray):
            return obj.tolist()

        if isinstance(obj, np.int64):
            return int(obj)

        if isinstance(obj, Parametrization):
            return type(obj).__name__

        if isinstance(obj, SharedParams):
            return obj.__dict__

        if isinstance(obj, Covariate):
            # Only include 'param' and 'values' to avoid cluttering
            # the config file.
            return dict(param=obj.param, values=obj.values)

        # there were serialization problems with this object
        if isinstance(obj, LbfgsInvHessProduct):
            return str(obj)

        return json.JSONEncoder.default(self, obj)



"""
SFS utilities.
"""

__author__ = "Janek Sendrowski"
__contact__ = "sendrowski.janek@gmail.com"
__date__ = "2022-07-24"

import logging
from typing import Dict, List, Union, Iterable, Any, Literal, Sequence

import numpy as np
import pandas as pd
from scipy.stats import hypergeom

from .io_handlers import download_if_url

# get logger
logger = logging.getLogger('fastdfe')


def standard_kingman(n: int) -> 'Spectrum':
    """
    Get standard Kingman SFS for theta = 1.

    :param n: Standard Kingman SFS
    :return: Spectrum
    """
    return Spectrum(pad(1 / np.arange(1, int(n))))


def pad(counts: Sequence) -> np.ndarray:
    """
    Pad array with monomorphic counts.

    :param counts: SFS counts to pad
    :return: Padded array
    """
    return np.array([0] + list(counts) + [0])


class Spectrum(Iterable):
    """
    Class for holding and manipulating a site-frequency spectrum.
    """

    def __init__(self, data: Sequence[float]):
        """
        Initialize spectrum.

        :param data: SFS counts
        """
        self.data: np.ndarray = np.array(data, dtype=float)

    @property
    def n(self) -> int:
        """
        The sample size.

        :return: Sample size
        """
        return self.data.shape[0] - 1

    @property
    def n_sites(self) -> float:
        """
        The total number of sites.

        :return: Total number of sites
        """
        return sum(self.data)

    @property
    def n_div(self) -> float:
        """
        Number of divergence counts.

        :return: Number of divergence counts
        """
        return self.data[-1]

    @property
    def has_div(self) -> bool:
        """
        Whether n_div was specified.

        :return: Whether n_div was specified
        """
        return self.n_div != 0

    @property
    def n_monomorphic(self) -> float:
        """
        Number of monomorphic sites.

        :return: Number of monomorphic sites
        """
        return self.data[0] + self.data[-1]

    @property
    def polymorphic(self) -> np.ndarray:
        """
        Get the polymorphic counts.

        :return: Polymorphic counts
        """
        return self.data[1:-1]

    @property
    def n_polymorphic(self) -> np.ndarray:
        """
        Get the polymorphic counts.

        :return: Polymorphic counts
        """
        return np.sum(self.polymorphic)

    def to_list(self) -> list:
        """
        Convert to list.

        :return: SFS counts
        """
        return list(self.data)

    def to_spectra(self) -> 'Spectra':
        """
        Convert to Spectra object.

        :return: Spectra object
        """
        return Spectra.from_spectrum(self)

    def to_file(self, file: str):
        """
        Save object to file.

        :param file: File name
        """
        self.to_spectra().to_file(file)

    @staticmethod
    def from_file(file: str) -> 'Spectrum':
        """
        Load object from file.

        :param file: File name
        :return: Spectrum object
        """
        return Spectra.from_file(file).to_spectrum()

    def to_numpy(self) -> np.ndarray:
        """
        Convert to array.

        :return: SFS counts
        """
        return self.data

    @property
    def theta(self) -> float:
        """
        Calculate site-wise population mutation rate using Watterson's estimator.
        Note that theta is given per site, i.e. Watterson's estimator is divided by the
        total number of sites (:attr:`n_sites`).
        """
        return self.Theta / self.n_sites

    @property
    def Theta(self) -> float:
        """
        Calculate population mutation rate using Watterson's estimator.

        .. note:: Property :attr:`Theta` is not normalized by the total number of sites, unlike :attr:`theta`.
        """
        return self.n_polymorphic / np.sum(1 / np.arange(1, self.n))

    def fold(self) -> 'Spectrum':
        """
        Fold the site-frequency spectrum.

        :return: Folded spectrum
        """
        mid = (self.n + 1) // 2
        data = self.data.copy()

        data[:mid] += data[-mid:][::-1]
        data[-mid:] = 0

        return Spectrum(data)

    def misidentify(self, epsilon: float) -> 'Spectrum':
        """
        Introduce ancestral misidentification at rate epsilon. Note that monomorphic counts won't be affected.

        :param epsilon: Misidentification rate (0 <= epsilon <= 1)
        :return: Spectrum with misidentification applied
        :raise ValueError: If epsilon is not between 0 and 1
        """
        if not 0 <= epsilon <= 1:
            raise ValueError("epsilon must be between 0 and 1")

        data = self.data.copy()
        n = self.n

        flipped = epsilon * data[1:n][::-1]
        retained = (1 - epsilon) * data[1:n]
        data[1:n] = retained + flipped

        return Spectrum(data)

    def subsample(
            self,
            n: int,
            mode: Literal['random', 'probabilistic'] = 'probabilistic',
            seed: int | None = None
    ) -> 'Spectrum':
        """
        Subsample spectrum to a given sample size.

        .. warning::
            If using the 'random' mode, The SFS counts are cast to integers before subsampling so this will
            only provide sensible results if the SFS counts are integers or if they are large enough to be
            approximated well by integers. The 'probabilistic' mode does not have this limitation.

        :param n: Sample size
        :param mode: Subsampling mode. Either 'random' or 'probabilistic'.
        :param seed: Seed for random number generator. Only for 'random' mode.
        :return: Subsampled spectrum
        """
        if n >= self.n:
            raise ValueError(f'Subsampled sample size {n} must be smaller than original sample size {self.n}.')

        if mode not in ['random', 'probabilistic']:
            raise ValueError(f'Unknown subsampling mode {mode}.')

        subsample = np.zeros(n + 1, dtype=float)

        if mode == 'random':
            # add monomorphic counts
            subsample[0] = self.data[0]
            subsample[-1] = self.data[-1]

            # iterate over spectrum and subsample hypergeometrically
            for i, m in enumerate(self.polymorphic.astype(int)):
                # get subsampled counts
                samples = hypergeom.rvs(M=self.n, n=i + 1, N=n, size=m, random_state=seed)

                # add subsampled counts
                subsample += np.histogram(samples, bins=np.arange(n + 2))[0]
        else:
            for i, m in enumerate(self.data):
                probs = hypergeom.pmf(k=range(n + 1), M=self.n, n=i, N=n)

                # add subsampled counts
                subsample += m * probs

        sfs = Spectrum(subsample)

        # fold if original spectrum was folded
        if self.is_folded():
            sfs = sfs.fold()

        return sfs

    def resample(self, seed: int = None) -> 'Spectrum':
        """
        Resample SFS assuming independent Poisson counts.

        :param seed: Seed for random number generator.
        :return: Resampled spectrum.
        """
        seed = None if seed is None else int(seed)

        return Spectrum.from_polydfe(
            # resample polymorphic sites only
            polymorphic=np.random.default_rng(seed=seed).poisson(lam=self.polymorphic),
            n_sites=self.n_sites,
            n_div=self.n_div
        )

    def is_folded(self) -> bool:
        """
        Check if the site-frequency spectrum is folded.

        :return: True if folded, False otherwise
        """
        mid = (self.n + 1) // 2

        return np.all(self.data[-mid:] == 0)

    def normalize(self) -> 'Spectrum':
        """
        Normalize SFS so that all non-monomorphic counts add up to 1.

        :return: Normalized spectrum
        """
        # copy array
        data = self.data.copy()

        # normalize counts
        data[1:-1] /= data[1:-1].sum()

        return Spectrum(data)

    def copy(self) -> 'Spectrum':
        """
        Copy the spectrum.

        :return: Copy of the spectrum
        """
        return Spectrum(self.data.copy())

    @staticmethod
    def from_polymorphic(data: Sequence) -> 'Spectrum':
        """
        Create Spectrum from polymorphic counts only.

        :param data: Polymorphic counts
        :return: Spectrum
        """
        return Spectrum([0] + list(data) + [0])

    @staticmethod
    def from_list(data: Sequence) -> 'Spectrum':
        """
        Create Spectrum from list.

        :param data: SFS counts
        :return: Spectrum
        """
        return Spectrum(data)

    @staticmethod
    def from_polydfe(
            polymorphic: Sequence,
            n_sites: float,
            n_div: float
    ) -> 'Spectrum':
        """
        Create Spectra from polyDFE specification which treats the number
        of mutational target sites and the divergence counts separately.

        :param polymorphic: Polymorphic counts
        :param n_sites: Total number of sites
        :param n_div: Number of divergence counts
        :return: Spectrum
        """
        # determine number of monomorphic ancestral counts
        n_monomorphic = n_sites - np.sum(list(polymorphic) + [n_div])

        data = [n_monomorphic] + list(polymorphic) + [n_div]

        return Spectrum(data)

    @staticmethod
    def _array_or_scalar(data: Iterable | float) -> np.ndarray | float:
        """
        Convert to array if iterable or return scalar otherwise.

        :param data: Iterable or scalar.
        :return: Array or scalar
        """
        if isinstance(data, Iterable):
            return np.array(list(data))

        return data

    def __mul__(self, other: Iterable | float) -> 'Spectrum':
        """
        Multiply spectrum.

        :param other: Iterable or scalar
        :return: Spectrum
        """
        return Spectrum(self.data * self._array_or_scalar(other))

    __rmul__ = __mul__

    def __add__(self, other: Iterable | float) -> 'Spectrum':
        """
        Add spectrum.

        :param other: Iterable or scalar
        :return: Spectrum
        """
        return Spectrum(self.data + self._array_or_scalar(other))

    def __sub__(self, other: Iterable | float) -> 'Spectrum':
        """
        Subtract spectrum.

        :param other: Iterable or scalar
        :return: Spectrum
        """
        return Spectrum(self.data - self._array_or_scalar(other))

    def __pow__(self, other: Iterable | float) -> 'Spectrum':
        """
        Power operator.

        :param other: Iterable or scalar
        :return: Spectrum
        """
        return Spectrum(self.data ** self._array_or_scalar(other))

    def __floordiv__(self, other: Iterable | float) -> 'Spectrum':
        """
        Divide spectrum.

        :param other: Iterable or scalar
        :return: Spectrum
        """
        return Spectrum(self.data // self._array_or_scalar(other))

    def __truediv__(self, other: Iterable | float) -> 'Spectrum':
        """
        Add spectrum.

        :param other: Iterable or scalar
        :return: Spectrum
        """
        return Spectrum(self.data / self._array_or_scalar(other))

    def __iter__(self):
        """
        Get iterator.

        :return: Iterator
        """
        return self.data.__iter__()

    def plot(
            self,
            show: bool = True,
            file: str = None,
            title: str = None,
            log_scale: bool = False,
            show_monomorphic: bool = False,
            kwargs_legend: dict = dict(prop=dict(size=8)),
            ax: 'plt.Axes' = None
    ) -> 'plt.Axes':
        """
        Plot spectrum.

        :param show: Whether to show plot.
        :param file: File to save plot to.
        :param title: Title of plot.
        :param log_scale: Whether to use log scale on y-axis.
        :param show_monomorphic: Whether to show monomorphic counts.
        :param kwargs_legend: Keyword arguments passed to :meth:`plt.legend`. Only for Python visualization backend.
        :param ax: Axes to plot on. Only for Python visualization backend.
        :return: Axes
        """
        from .visualization import Visualization

        return Visualization.plot_spectra(
            spectra=[self.to_list()],
            file=file,
            show=show,
            title=title,
            log_scale=log_scale,
            show_monomorphic=show_monomorphic,
            ax=ax,
            kwargs_legend=kwargs_legend
        )

    @staticmethod
    def standard_kingman(n: int, n_monomorphic: int = 0) -> 'Spectrum':
        """
        Get standard Kingman SFS.

        :param n: sample size
        :param n_monomorphic: Number of monomorphic sites.
        :return: Standard Kingman SFS
        """
        sfs = standard_kingman(n)
        sfs.data[0] = n_monomorphic

        return sfs


class Spectra:
    """
    Class for holding and manipulating site-frequency spectra of multiple types.
    """

    def __init__(self, data: Dict[str, Iterable]):
        """
        Initialize spectra.

        :param data: Dictionary of SFS counts keyed by type
        """
        self.data: pd.DataFrame = pd.DataFrame(data)

    @property
    def n(self) -> int:
        """
        The sample size.

        :return: Sample size
        """
        return self.data.shape[0] - 1

    @property
    def k(self) -> int:
        """
        The number of types.

        :return: Number of types
        """
        return self.data.shape[1]

    @property
    def n_monomorphic(self) -> pd.Series:
        """
        The number of monomorphic sites.

        :return: Number of monomorphic sites
        """
        return self.data.iloc[0] + self.data.iloc[-1]

    @property
    def polymorphic(self) -> np.ndarray:
        """
        The polymorphic counts.

        :return: Polymorphic counts
        """
        return self.data[1:-1]

    @property
    def n_polymorphic(self) -> np.ndarray:
        """
        The total number of polymorphic counts.

        :return: Total number of polymorphic counts for each type
        """
        return self.polymorphic.sum()

    @staticmethod
    def from_list(data: Sequence, types: List) -> 'Spectra':
        """
        Create from array of spectra.
        Note that data.ndim needs to be 2.

        :param data: Array of spectra
        :param types: Types
        :return: Spectra
        """
        return Spectra(dict((t, d) for t, d in zip(types, data)))

    @property
    def types(self) -> List[str]:
        """
        The types.

        :return: Types
        """
        return self.data.columns.to_list()

    @property
    def n_sites(self) -> pd.Series:
        """
        The number of mutational target sites which is the sum of all SFS entries.

        :return: Number of mutational target sites for each type
        """
        return self.data.sum()

    @property
    def n_div(self) -> pd.Series:
        """
        The number of divergence counts.

        :return: Number of divergence counts for each type
        """
        return self.data.iloc[-1]

    @property
    def has_div(self) -> pd.Series:
        """
        Whether n_div was specified.

        :return: Whether n_div was specified for each type
        """
        # noinspection PyTypeChecker
        return self.n_div != 0

    @property
    def theta(self) -> pd.Series:
        """
        Calculate site-wise population mutation rate using Watterson's estimator.
        Note that theta is given per site, i.e. Watterson's estimator is divided by the
        total number of sites (:attr:`n_sites`).
        """
        return self.Theta / self.n_sites

    @property
    def Theta(self) -> pd.Series:
        """
        Calculate population mutation rate using Watterson's estimator.

        .. note:: Property :attr:`Theta` is not normalized by the total number of sites, unlike :attr:`theta`.
        """
        return self.n_polymorphic / np.sum(1 / np.arange(1, self.n))

    def normalize(self) -> 'Spectra':
        """
        Normalize spectra by sum of all entries.

        :return: Normalized spectra
        """
        return self / self.data.sum()

    def to_file(self, file: str):
        """
        Save object to file.

        :param file: File name
        """
        self.data.to_csv(file, index=False)

    def to_spectra(self) -> Dict[str, Spectrum]:
        """
        Convert to dictionary of spectrum objects.

        :return: Dictionary of spectrum objects
        """
        return dict((t, self.select(t, use_regex=False)) for t in self.types)

    def to_dataframe(self) -> pd.DataFrame:
        """
        Get representation as dataframe.

        :return: Dataframe
        """
        return self.data

    def to_numpy(self) -> np.ndarray:
        """
        Convert to numpy array.

        :return: Numpy array
        """
        return self.data.to_numpy().T

    def to_list(self) -> list:
        """
        Convert to nested list.

        :return: Nested list
        """
        return list(list(d) for d in self.to_numpy())

    def to_dict(self) -> dict:
        """
        Convert to dictionary.

        :return: Dictionary of lists
        """
        # return dictionary of lists
        return dict((k, list(v.values())) for k, v in self.data.to_dict().items())

    def __mul__(self, other: Any) -> 'Spectra':
        """
        Multiply Spectra.

        :param other: Scalar
        :return: Spectra
        """
        return Spectra.from_dataframe(self.data * other)

    __rmul__ = __mul__

    def __floordiv__(self, other: Any) -> 'Spectra':
        """
        Divide Spectra.

        :param other: Scalar
        :return: Spectra
        """
        return Spectra.from_dataframe(self.data // other)

    def __truediv__(self, other: Any) -> 'Spectra':
        """
        Divide Spectra.

        :param other: Scalar
        :return: Spectra
        """
        return Spectra.from_dataframe(self.data / other)

    def __len__(self) -> int:
        """
        Get number of spectra.

        :return: Number of spectra
        """
        return self.k

    def __add__(self, other: 'Spectra') -> 'Spectra':
        """
        Merge types of two spectra objects by adding up their counts entry-wise.

        :param other: Spectra object
        :return: Spectra with merged types
        """
        return Spectra.from_dataframe(self.data.add(other.data, fill_value=0))

    def __getitem__(
            self,
            keys: str | List[str] | np.ndarray | tuple,
            use_regex: bool = True
    ) -> Union['Spectrum', 'Spectra']:
        """
        Get item.

        :param keys: String or list of strings, possibly regex to match type names
        :param use_regex: Whether to use regex to match type names
        :return: Spectrum or Spectra object depending on the number of matches
        """
        # whether the input in an array
        is_array = isinstance(keys, (np.ndarray, list, tuple))

        if use_regex:
            # subset dataframe using column names using regex
            subset = self.data.loc[:, self.data.columns.str.fullmatch('|'.join(keys) if is_array else keys)]
        else:
            # subset dataframe using column names
            subset = self.data.loc[:, keys]

        # return spectrum object if we have a series
        if isinstance(subset, pd.Series):
            return Spectrum(list(subset))

        # return spectrum object if only one column is left
        # and if not multiple keys were supplied
        if subset.shape[1] == 1 and not is_array:
            return Spectrum(list(subset.iloc[:, 0]))

        # wrap subset dataframe in spectra object
        return Spectra.from_dataframe(subset)

    def __setitem__(self, key: str, s: Spectrum):
        """
        Save new spectrum as type.

        :param key: Type
        :param s: Spectrum
        """
        self.data[key] = s.to_list()

    def __iter__(self):
        """
        Get iterator.

        :return: Iterator
        """
        return self.data.__iter__()

    def select(
            self,
            keys: str | List[str] | np.ndarray | tuple,
            use_regex: bool = True
    ) -> 'Spectra':
        """
        Select types. Alias for __getitem__.

        :param keys: String or list of strings, possibly regex to match type names
        :param use_regex: Whether to use regex to match type names
        :return: Spectrum or Spectra depending on the number of matches
        """
        return self.__getitem__(keys, use_regex=use_regex)

    def copy(self) -> 'Spectra':
        """
        Copy object.

        :return: Copy of object
        """
        return Spectra.from_dataframe(self.data.copy())

    def _to_multi_index(self) -> 'Spectra':
        """
        Convert to Spectra object with multi-indexed columns.

        :return: Spectra object with multi-indexed columns
        """
        other = self.copy()
        columns = [tuple(col.split('.')) for col in other.data.columns]
        other.data.columns = pd.MultiIndex.from_tuples(columns)

        return other

    def _to_single_index(self) -> 'Spectra':
        """
        Convert to Spectra object with single-indexed columns (using dot notation).

        :return: Spectra object with single-indexed columns
        """
        other = self.copy()

        if other.data.columns.nlevels > 1:
            columns = other.data.columns.map('.'.join)
            other.data.columns = columns

        return other

    def get_empty(self) -> 'Spectra':
        """
        Get a Spectra object with zero counts but having the same shape and types as self.

        :return: Spectra object with zero counts
        """
        return Spectra.from_dataframe(pd.DataFrame(0, index=self.data.index, columns=self.data.columns))

    def merge_groups(self, level: List[int] | int = 0) -> 'Spectra':
        """
        Group over given levels and sum up spectra so the spectra
        are summed over the levels that were not specified.

        :param level: Level(s) to group over
        :return: Spectra object with merged groups
        """
        # cast to int
        level = [int(l) for l in level] if isinstance(level, Iterable) else int(level)

        return Spectra.from_dataframe(self._to_multi_index().data.T.groupby(level=level).sum().T)._to_single_index()

    def has_dots(self) -> bool:
        """
        Check whether column names contain dots.

        :return: True if column names contain dots, False otherwise
        """
        return any('.' in col for col in self.data.columns)

    def replace_dots(self, replacement: str = '_') -> 'Spectra':
        """
        Replace dots in column names with a given string.

        :param replacement: Replacement string
        :return: Spectra object with replaced dots
        """
        other = self.copy()
        other.data.columns = other.data.columns.str.replace('.', replacement)

        return other

    @property
    def all(self) -> 'Spectrum':
        """
        The 'all' type equals the sum of all spectra.

        :return: Spectrum object
        """
        return Spectrum(self.data.sum(axis=1).to_list())

    def combine(self, s: 'Spectra') -> 'Spectra':
        """
        Merge types of two Spectra objects.

        :param s: Other Spectra object
        :return: Merged Spectra object
        """
        return Spectra(self.to_dict() | s.to_dict())

    @staticmethod
    def from_dict(data: dict) -> 'Spectra':
        """
        Load from nested dictionary first indexed by types and then by samples.

        :param data: Dictionary of lists indexed by types
        :return: Spectra object
        """
        lists = [list(v.values() if isinstance(v, dict) else v) for v in data.values()]

        return Spectra.from_list(lists, types=list(data.keys()))

    @staticmethod
    def from_dataframe(data: pd.DataFrame) -> 'Spectra':
        """
        Load Spectra object from dataframe.

        :param data: Dataframe
        :return: Spectra object
        """
        return Spectra.from_dict(data.to_dict())

    @classmethod
    def from_file(cls, file: str) -> 'Spectra':
        """
        Save object to file.

        :param file: Path to file, possibly URL
        :return: Spectra object
        """
        return Spectra.from_dataframe(pd.read_csv(download_if_url(
            file,
            desc=f'{cls.__name__}>Downloading file'))
        )

    @staticmethod
    def from_spectra(spectra: Dict[str, Spectrum]) -> 'Spectra':
        """
        Create from dict of spectrum objects indexed by type.

        :param spectra: Dictionary of spectrum objects indexed by type
        :return: Spectra object
        """
        return Spectra.from_list([sfs.to_list() for sfs in spectra.values()], types=list(spectra.keys()))

    @staticmethod
    def from_spectrum(sfs: Spectrum) -> 'Spectra':
        """
        Create from single spectrum object. The type of the spectrum is set to 'all'.

        :param sfs: Spectrum
        :return: Spectra object
        """
        return Spectra.from_spectra(dict(all=sfs))

    def to_spectrum(self) -> Spectrum:
        """
        Convert to Spectrum object by summing over all types.

        :return: Spectrum object
        """
        return self.all

    def plot(
            self,
            show: bool = True,
            file: str = None,
            title: str = None,
            log_scale: bool = False,
            use_subplots: bool = False,
            show_monomorphic: bool = False,
            kwargs_legend: dict = dict(prop=dict(size=8)),
            ax: 'plt.Axes' = None
    ) -> 'plt.Axes':
        """
        Visualize spectra.

        :param show: Whether to show the plot.
        :param file: File name to save the plot to.
        :param title: Plot title.
        :param log_scale: Whether to use log scale on y-axis.
        :param use_subplots: Whether to use subplots. Only for Python visualization backend.
        :param show_monomorphic: Whether to show monomorphic sites.
        :param kwargs_legend: Keyword arguments passed to :meth:`plt.legend`. Only for Python visualization backend.
        :param ax: Axes to plot on. Only for Python visualization backend and if ``use_subplots`` is ``False``.
        :return: Axes
        """
        from .visualization import Visualization

        return Visualization.plot_spectra(
            spectra=list(list(v) for v in self.to_spectra().values()),
            labels=self.types,
            file=file,
            show=show,
            title=title,
            log_scale=log_scale,
            use_subplots=use_subplots,
            show_monomorphic=show_monomorphic,
            kwargs_legend=kwargs_legend,
            ax=ax
        )

    def drop_empty(self) -> 'Spectra':
        """
        Remove types whose spectra have no counts.

        :return: Spectra with non-empty types
        """
        return Spectra.from_dataframe(self.data.loc[:, self.data.any()])

    def drop_zero_entries(self) -> 'Spectra':
        """
        Remove types whose spectra have some zero entries.
        Note that we ignore zero counts in the last entry i.e. fixed derived alleles.

        :return: Spectra with non-zero entries
        """
        return Spectra.from_dataframe(self.data.loc[:, self.data[:-1].all()])

    def drop_sparse(self, n_polymorphic: int) -> 'Spectra':
        """
        Remove types whose spectra have fewer than equal ``n_polymorphic`` polymorphic sites.

        :return: Spectra
        """
        return Spectra.from_dataframe(self.data.loc[:, self.data[1:-1].sum() > int(n_polymorphic)])

    def rename(self, names: List[str]) -> 'Spectra':
        """
        Rename types.

        :param names: New names
        :return: Spectra with renamed types
        """
        other = self.copy()
        other.data.columns = names

        return other

    def prefix(self, prefix: str) -> 'Spectra':
        """
        Prefix types, i.e. 'type' -> 'prefix.type' for all types.

        :param prefix: Prefix
        :return: Spectra with prefixed types
        """
        return self.rename([prefix + '.' + col for col in self.types])

    def reorder_levels(self, levels: List[int]) -> 'Spectra':
        """
        Reorder levels.

        :param levels: New order of levels
        :return: Spectra with reordered levels
        """
        s = self._to_multi_index()
        s.data.columns = s.data.columns.reorder_levels(levels)
        s = s._to_single_index()

        return s

    def print(self):
        """
        Print spectra.
        """
        print(self.data.T)

    def fold(self):
        """
        Fold spectra.

        :return: Folded spectra
        """
        return Spectra.from_spectra({t: s.fold() for t, s in self.to_spectra().items()})

    def subsample(
            self,
            n: int,
            mode: Literal['random', 'probabilistic'] = 'probabilistic'
    ) -> 'Spectra':
        """
        Subsample spectra to a given sample size.

        .. warning::
            If using the 'random' mode, The SFS counts are cast to integers before subsampling so this will
            only provide sensible results if the SFS counts are integers or if they are large enough to be
            approximated by integers. The 'probabilistic' mode does not have this limitation.

        :param n: Sample size
        :param mode: Subsampling mode. Either 'random' or 'probabilistic'.
        :return: Subsampled spectra
        """
        return Spectra.from_spectra({t: s.subsample(n, mode) for t, s in self.to_spectra().items()})

    def is_folded(self) -> Dict[str, bool]:
        """
        Check whether spectra are folded.

        :return: Dictionary of types and whether they are folded
        """
        return {t: s.is_folded() for t, s in self.to_spectra().items()}

    def sort_types(self) -> 'Spectra':
        """
        Sort types alphabetically.

        :return: Sorted spectra object
        """
        return Spectra.from_dataframe(self.data.sort_index(axis=1))


def parse_polydfe_sfs_config(file: str) -> Spectra:
    """
    Parse frequency spectra and mutational target site from
    polyDFE configuration file.

    :param file: File name
    :return: Spectra object
    """
    df = pd.read_csv(file, header=None, comment='#')

    # parse number of spectra and sample size
    n_neut, n_sel, n = np.array(df.iloc[0][0].split()).astype(int)

    # issue notice about number of spectra and sample size
    logger.info(f'Parsing {n_neut} neutral and {n_sel} selected SFS with '
                f'a sample size of {n}.')

    # issue notice that fastDFE does not support variable mutation rates
    if n_neut > 1 or n_sel > 1:
        logger.info('Note that fastDFE does not model variable mutation rates '
                    'as this did not turn out to change the inference result. '
                    'The parsed spectra are thus merged together.')

    def to_spectrum(data: np.ndarray) -> Spectrum:
        """
        Parse spectrum and number of mutational target sites.
        We ignore the number of mutational target sites for divergence counts
        but include the divergence counts for completeness of the SFS.

        :param data: Spectrum data
        :return: Spectrum object
        """
        # iterate over spectra and merge them as we do not
        # support variable mutation rates
        data_merged = data.sum(axis=0)

        # polymorphic counts
        polymorphic = list(data_merged[:n - 1])

        # parse number mutational target sites for ingroup
        n_sites = float(data_merged[n - 1])

        # parse optional divergence counts
        n_div = float(data_merged[n]) if n < data_merged.shape[0] else 0

        return Spectrum.from_polydfe(polymorphic, n_sites=n_sites, n_div=n_div)

    # iterate over spectra and merge them as we do not
    # support variable mutation rates
    data_neut = np.array([df.iloc[i][0].split() for i in range(1, n_neut + 1)], dtype=float)
    sfs_neut = to_spectrum(data_neut)

    # iterate over spectra and merge them as we do not
    # support variable mutation rates
    data_sel = np.array([df.iloc[i][0].split() for i in range(n_neut + 1, n_neut + n_sel + 1)], dtype=float)
    sfs_sel = to_spectrum(data_sel)

    return Spectra.from_spectra(dict(
        sfs_neut=sfs_neut,
        sfs_sel=sfs_sel
    ))



"""
VCF filtrations and a filterer to apply them.
"""

__author__ = "Janek Sendrowski"
__contact__ = "sendrowski.janek@gmail.com"
__date__ = "2023-05-11"

import functools
import logging
from abc import ABC, abstractmethod
from typing import Iterable, List, Optional, Callable, Dict, Union

import numpy as np
import pandas as pd

from .annotation import DegeneracyAnnotation
from .io_handlers import get_major_base, MultiHandler, get_called_bases, DummyVariant

# get logger
logger = logging.getLogger('fastdfe')


def _count_filtered(func: Callable) -> Callable:
    """
    Decorator that increases ``self.n_filtered`` by 1 if the decorated function returns False.
    """

    @functools.wraps(func)
    def wrapper(self, variant):
        """
        Wrapper function.

        :param self: Self.
        :param variant: The variant to filter.
        :return: The result of the decorated function.
        """
        result = func(self, variant)
        if not result:
            self.n_filtered += 1
        return result

    return wrapper


class Filtration(ABC):
    """
    Base class for filtering sites based on certain criteria.
    """

    #: The number of sites that didn't pass the filter.
    n_filtered: int = 0

    def __init__(self):
        """
        Initialize filtration.
        """
        #: The logger.
        self._logger = logger.getChild(self.__class__.__name__)

        #: The handler.
        self._handler: MultiHandler | None = None

    @abstractmethod
    @_count_filtered
    def filter_site(self, variant: Union['cyvcf2.Variant', DummyVariant]) -> bool:
        """
        Filter site.

        :param variant: The variant to filter.
        :return: ``True`` if the variant should be kept, ``False`` otherwise.
        """
        pass

    def _setup(self, handler: MultiHandler):
        """
        Perform any necessary pre-processing. This method is called before the actual filtration.

        :param handler: The handler.
        """
        self._handler = handler

    def _rewind(self):
        """
        Rewind the filtration.
        """
        self.n_filtered = 0

    def _teardown(self):
        """
        Perform any necessary post-processing. This method is called after the actual filtration.
        """
        self._logger.info(f"Filtered out {self.n_filtered} sites.")


class MaskedFiltration(Filtration, ABC):
    """
    Filter sites based on a samples mask.
    """

    def __init__(
            self,
            use_parser: bool = True,
            include_samples: List[str] | None = None,
            exclude_samples: List[str] | None = None
    ):
        """
        Create a new filtration instance.

        :param use_parser: Whether to use the samples mask from the parser, if used together with parser.
        :param include_samples: The samples to include, defaults to all samples.
        :param exclude_samples: The samples to exclude, defaults to no samples.
        """
        super().__init__()

        #: Whether to use the samples mask from the parser, if used together with parser.
        self.use_parser: bool = use_parser

        #: The samples to include.
        self.include_samples: List[str] | None = include_samples

        #: The samples to exclude.
        self.exclude_samples: List[str] | None = exclude_samples

        #: The samples mask.
        self._samples_mask: np.ndarray | None = None

    def _prepare_samples_mask(self):
        """
        Prepare the samples mask.
        """
        from .parser import Parser

        if self.use_parser and isinstance(self._handler, Parser):

            # use samples mask from parser
            self._samples_mask = self._handler._samples_mask

        elif self.include_samples is None and self.exclude_samples is None:

            # no samples mask
            self._samples_mask = None

        else:

            # determine samples to include
            if self.include_samples is None:

                mask = np.ones(len(self._handler._reader.samples)).astype(bool)
            else:
                mask = np.isin(self._handler._reader.samples, self.include_samples)

            # determine samples to exclude
            if self.exclude_samples is not None:
                mask &= ~np.isin(self._handler._reader.samples, self.exclude_samples)

            self._samples_mask = mask

    def _setup(self, handler: MultiHandler):
        """
        Prepare the samples mask.

        :param handler: The handler.
        """
        super()._setup(handler)

        # prepare samples mask
        self._prepare_samples_mask()


class SNPFiltration(MaskedFiltration):
    """
    Only keep SNPs. Note that this entails discarding mono-morphic sites.
    """

    @_count_filtered
    def filter_site(self, variant: Union['cyvcf2.Variant', DummyVariant]) -> bool:
        """
        Filter site.

        :param variant: The variant to filter.
        :return: ``True`` if the variant is an SNP, ``False`` otherwise.
        """
        # simply check whether the variant is an SNP if we don't have a samples mask
        if self._samples_mask is None or isinstance(variant, DummyVariant):
            return variant.is_snp

        # otherwise check whether the variant is an SNP among the included samples
        return len(np.unique(get_called_bases(variant.gt_bases[self._samples_mask]))) > 1


class SNVFiltration(Filtration):
    """
    Only keep single site variants (discard indels and MNPs but keep monomorphic sites).
    """

    @_count_filtered
    def filter_site(self, variant: Union['cyvcf2.Variant', DummyVariant]) -> bool:
        """
        Filter site.

        :param variant: The variant to filter.
        :return: ``True`` if the variant is kept, ``False`` otherwise.
        """
        return np.all([alt in ['A', 'C', 'G', 'T'] for alt in [variant.REF] + variant.ALT])


class PolyAllelicFiltration(MaskedFiltration):
    """
    Filter out poly-allelic sites.
    """

    @_count_filtered
    def filter_site(self, variant: Union['cyvcf2.Variant', DummyVariant]) -> bool:
        """
        Filter site. Note that we don't check explicitly all alleles, but rather
        rely on ``ALT`` field.

        :param variant: The variant to filter.
        :return: ``True`` if the variant is not poly-allelic, ``False`` otherwise.
        """
        # if we don't have a samples mask, simply check whether the variant is poly-allelic
        if self._samples_mask is None or isinstance(variant, DummyVariant):
            return len(variant.ALT) < 2

        # otherwise check whether the variant is poly-allelic among the included samples
        return len(np.unique(get_called_bases(variant.gt_bases[self._samples_mask]))) < 3


class AllFiltration(Filtration):
    """
    Filter out all sites. Only useful for testing purposes.
    """

    @_count_filtered
    def filter_site(self, variant: Union['cyvcf2.Variant', DummyVariant]) -> bool:
        """
        Filter site.

        :param variant: The variant to filter.
        :return: ``False``.
        """
        return False


class NoFiltration(Filtration):
    """
    Do not filter out any sites. Only useful for testing purposes.
    """

    @_count_filtered
    def filter_site(self, variant: Union['cyvcf2.Variant', DummyVariant]) -> bool:
        """
        Filter site.

        :param variant: The variant to filter.
        :return: ``True``.
        """
        return True


class CodingSequenceFiltration(Filtration):
    """
    Filter out sites that are not in coding sequences. This filter should find frequent use when parsing
    spectra for DFE inference as we only consider sites in coding sequences for this purpose.
    By using it, the annotation and parsing of unnecessary sites can be avoided which increases the speed.
    Note that we assume here that within contigs, sites in the GFF file are sorted by position in ascending order.

    For this filtration to work, we require a GFF file (passed to :class:`~fastdfe.parser.Parser` or
    :class:`~fastdfe.filtration.Filterer`).
    """

    def __init__(self):
        """
        Create a new filtration instance.
        """
        Filtration.__init__(self)

        #: The coding sequence enclosing the current variant or the closest one downstream.
        self.cd: Optional[pd.Series] = None

        #: The number of processed sites.
        self.n_processed: int = 0

    def _setup(self, handler: MultiHandler):
        """
        Touch the GFF file to load it.

        :param handler: The handler.
        """
        # require GFF file
        handler._require_gff(self.__class__.__name__)

        # setup GFF handler
        super()._setup(handler)

        # load coding sequences
        _ = handler._cds

    def _rewind(self):
        """
        Rewind the filtration.
        """
        super()._rewind()

        # reset coding sequence
        self.cd = None

    @_count_filtered
    def filter_site(self, v: Union['cyvcf2.Variant', DummyVariant]) -> bool:
        """
        Filter site by whether it is in a coding sequence.

        :param v: The variant to filter.
        :return: ``True`` if the variant is in a coding sequence, ``False`` otherwise.
        """
        aliases = self._handler.get_aliases(v.CHROM)

        # if self.cd is None or not on the same chromosome or ends before the variant
        if self.cd is None or self.cd.seqid not in aliases or v.POS > self.cd.end:

            # initialize mock coding sequence
            self.cd = pd.Series({
                'seqid': v.CHROM,
                'start': DegeneracyAnnotation._pos_mock,
                'end': DegeneracyAnnotation._pos_mock
            })

            # find coding sequences downstream
            cds = self._handler._cds[self._handler._cds['seqid'].isin(aliases) & (self._handler._cds['end'] >= v.POS)]

            if not cds.empty:
                # take the first coding sequence
                self.cd = cds.iloc[0]

                if self.cd.start == v.POS:
                    self._logger.debug(f'Found coding sequence for {v.CHROM}:{v.POS}.')
                else:
                    self._logger.debug(f'Found coding sequence downstream of {v.CHROM}:{v.POS}.')

            if self.n_processed == 0 and self.cd.start == DegeneracyAnnotation._pos_mock:
                self._logger.warning(f'No subsequent coding sequence found on the same contig as the first variant. '
                                     f'Please make sure this is the correct GFF file with contig names matching '
                                     f'the VCF file. You can use the aliases parameter to match contig names.')

        self.n_processed += 1

        # check whether the variant is in the current coding sequence
        if self.cd is not None and self.cd.seqid in aliases and self.cd.start <= v.POS <= self.cd.end:
            return True

        return False


class DeviantOutgroupFiltration(Filtration):
    """
    Filter out sites where the major allele of the specified outgroup samples differs from the major
    allele of the ingroup samples.
    """

    def __init__(
            self,
            outgroups: List[str],
            ingroups: List[str] = None,
            strict_mode: bool = True,
            retain_monomorphic: bool = True
    ):
        """
        Construct DeviantOutgroupFiltration.

        :param outgroups: The name of the outgroup samples to consider.
        :param ingroups: The name of the ingroup samples to consider, defaults to all samples but the outgroups.
        :param strict_mode: Whether to filter out sites where no outgroup sample is present, defaults to ``True``.
        :param retain_monomorphic: Whether to retain monomorphic sites, defaults to ``True``, which is faster.
        """
        super().__init__()

        #: The ingroup samples.
        self.ingroups: List[str] | None = ingroups

        #: The outgroup samples.
        self.outgroups: List[str] = outgroups

        #: Whether to filter out sites where no outgroup sample is present.
        self.strict_mode: bool = strict_mode

        #: Whether to retain monomorphic sites.
        self.retain_monomorphic: bool = retain_monomorphic

        #: The samples found in the VCF file.
        self.samples: Optional[np.ndarray] = None

        #: The ingroup mask.
        self.ingroup_mask: Optional[np.ndarray] = None

        #: The outgroup mask.
        self.outgroup_mask: Optional[np.ndarray] = None

    def _setup(self, handler: MultiHandler):
        """
        Touch the reader to load the samples.

        :param handler: The handler.
        """
        super()._setup(handler)

        # create samples array
        self.samples: np.ndarray = np.array(handler._reader.samples)

        # create ingroup and outgroup masks
        self._create_masks()

    def _create_masks(self):
        """
        Create ingroup and outgroup masks based on the samples.
        """

        # create outgroup masks
        self.outgroup_mask: np.ndarray = np.isin(self.samples, self.outgroups)

        # make sure all outgroups are present
        if self.outgroup_mask.sum() != len(self.outgroups):
            raise ValueError(f'Not all outgroup samples are present in the VCF file: {self.outgroups}')

        # create ingroup mask
        if self.ingroups is None:
            self.ingroup_mask = ~self.outgroup_mask
        else:
            self.ingroup_mask = np.isin(self.samples, self.ingroups)

    @_count_filtered
    def filter_site(self, variant: Union['cyvcf2.Variant', DummyVariant]) -> bool:
        """
        Filter site.

        :param variant: The variant to filter.
        :return: ``True`` if the variant should be kept, ``False`` otherwise.
        """
        # keep monomorphic sites if requested
        if not variant.is_snp and self.retain_monomorphic:
            return True

        # filter out dummies if retain_monomorphic is false
        if isinstance(variant, DummyVariant):
            return False

        # get major base among ingroup samples
        ingroup_base = get_major_base(variant.gt_bases[self.ingroup_mask])

        # get major base among outgroup samples
        outgroup_base = get_major_base(variant.gt_bases[self.outgroup_mask])

        # filter out if no outgroup base is present and strict mode is enabled
        if outgroup_base is None:
            return not self.strict_mode

        # filter out if outgroup base is different from ingroup base
        return ingroup_base == outgroup_base


class ExistingOutgroupFiltration(Filtration):
    """
    Filter out sites for which at least ``n_missing`` of the specified outgroup samples have no called base.
    """

    def __init__(self, outgroups: List[str], n_missing: int = 1):
        """
        Construct ExistingOutgroupFiltration.

        :param outgroups: The names of the outgroup samples considered.
        :param n_missing: The number of outgroup samples that need to be missing to fail the filter.
        """
        super().__init__()

        #: The outgroup samples.
        self.outgroups: List[str] = outgroups

        #: Minimum number of missing outgroups required to filter out a site.
        self.n_missing: int = n_missing

        #: The samples found in the VCF file.
        self.samples: Optional[np.ndarray] = None

        #: The outgroup mask.
        self.outgroup_mask: Optional[np.ndarray] = None

    def _setup(self, handler: MultiHandler):
        """
        Touch the reader to load the samples.

        :param handler: The handler.
        """
        super()._setup(handler)

        # create samples array
        self.samples: np.ndarray = np.array(handler._reader.samples)

        # create outgroup mask
        self._create_mask()

    def _create_mask(self):
        """
        Create outgroup mask based on the samples.
        """
        self.outgroup_mask: np.ndarray = np.isin(self.samples, self.outgroups)

    @_count_filtered
    def filter_site(self, variant: Union['cyvcf2.Variant', DummyVariant]) -> bool:
        """
        Filter site.

        :param variant: The variant to filter.
        :return: ``True`` if the variant should be kept, ``False`` otherwise.
        """
        # keep dummy variants
        if isinstance(variant, DummyVariant):
            return True

        # get outgroup genotypes
        outgroups = variant.gt_bases[self.outgroup_mask]

        # count how many outgroups have no called base
        missing_count = sum(len(get_called_bases(outgroup)) == 0 for outgroup in outgroups)

        # filter out if at least n outgroups are missing
        return missing_count < self.n_missing


class BiasedGCConversionFiltration(Filtration):
    """
    Only retain A<->T and G<->C substitutions (which are unaffected
    by biased gene conversion, see [CITGB]_).

    Mono-allelic sites are always retained, and we assume sites are at most bi-allelic. Note that the number of
    mutational target sites is reduced by this filtration.

    .. [CITGB] Pouyet et al., 'Background selection and biased
        gene conversion affect more than 95% of the human genome and bias demographic inferences.',
        Elife, 7:e36317, 2018
    """

    @_count_filtered
    def filter_site(self, variant: Union['cyvcf2.Variant', DummyVariant]) -> bool:
        """
        Remove bi-allelic sites that are not A<->T or G<->C substitutions.

        :param variant: The variant to filter.
        :return: ``True`` if the variant should be kept, ``False`` otherwise.
        """
        if variant.is_snp and len(variant.ALT) > 0:
            return (variant.REF, variant.ALT[0]) in [('A', 'T'), ('T', 'A'), ('G', 'C'), ('C', 'G')]

        return True

class ContigFiltration(Filtration):
    """
    Filter out sites that are not on the specified contigs.
    """

    def __init__(self, contigs: List[str]):
        """
        Construct ContigFiltration.

        :param contigs: The contigs to retain.
        """
        super().__init__()

        #: The contigs to retain.
        self.contigs: List[str] = contigs

    @_count_filtered
    def filter_site(self, variant: Union['cyvcf2.Variant', DummyVariant]) -> bool:
        """
        Filter site.

        :param variant: The variant to filter.
        :return: ``True`` if the variant is on one of the specified contigs, ``False`` otherwise.
        """
        return variant.CHROM in self.contigs


class Filterer(MultiHandler):
    """
    Filter a VCF file using a list of filtrations.

    Example usage:

    ::

        import fastdfe as fd

        # only keep variants in coding sequences
        f = fd.Filterer(
            vcf="http://ftp.1000genomes.ebi.ac.uk/vol1/ftp/data_collections/"
                "1000_genomes_project/release/20181203_biallelic_SNV/"
                "ALL.chr21.shapeit2_integrated_v1a.GRCh38.20181129.phased.vcf.gz",
            gff="http://ftp.ensembl.org/pub/release-109/gff3/homo_sapiens/"
                "Homo_sapiens.GRCh38.109.chromosome.21.gff3.gz",
            output='sapiens.chr21.coding.vcf.gz',
            filtrations=[fd.CodingSequenceFiltration()],
            aliases=dict(chr21=['21'])
        )

        f.filter()

    """

    def __init__(
            self,
            vcf: str | Iterable['cyvcf2.Variant'],
            output: str,
            gff: str | None = None,
            filtrations: List[Filtration] = [],
            info_ancestral: str = 'AA',
            max_sites: int = np.inf,
            seed: int | None = 0,
            cache: bool = True,
            aliases: Dict[str, List[str]] = {}
    ):
        """
        Create a new filter instance.

        :param vcf: The VCF file, possibly gzipped or a URL.
        :param output: The output file.
        :param gff: The GFF file, possibly gzipped or a URL. This argument is required for some filtrations.
        :param filtrations: The filtrations.
        :param info_ancestral: The info field for the ancestral allele.
        :param max_sites: The maximum number of sites to process.
        :param seed: The seed for the random number generator. Use ``None`` for no seed.
        :param cache: Whether to cache files downloaded from urls.
        :param aliases: Dictionary of aliases for the contigs in the VCF file, e.g. ``{'chr1': ['1']}``.
        """
        super().__init__(
            vcf=vcf,
            gff=gff,
            info_ancestral=info_ancestral,
            max_sites=max_sites,
            seed=seed,
            cache=cache,
            aliases=aliases
        )

        #: The filtrations.
        self.filtrations: List[Filtration] = filtrations

        #: The output file.
        self.output: str = output

        #: The number of sites that did not pass the filters.
        self.n_filtered: int = 0

        #: The VCF writer.
        self._writer: 'cyvcf2.Writer' | None = None

    def is_filtered(self, variant: Union['cyvcf2.Variant', DummyVariant]) -> bool:
        """
        Whether the given variant is kept.

        :param variant: The variant to check.
        :return: ``True`` if the variant is kept, ``False`` otherwise.
        """
        # filter the variant
        for filtration in self.filtrations:
            if not filtration.filter_site(variant):
                self.n_filtered += 1
                return False

        return True

    def _setup(self):
        """
        Set up the filtrations.
        """
        try:
            from cyvcf2 import Writer
        except ImportError:
            raise ImportError(
                "VCF support in fastdfe requires the optional 'cyvcf2' package. "
                "Please install fastdfe with the 'vcf' extra: pip install fastdfe[vcf]"
            )

        # setup filtrations
        for f in self.filtrations:
            f._setup(self)

        # create the writer
        self._writer = Writer(self.output, self._reader)

    def _teardown(self):
        """
        Tear down the filtrations.
        """
        for f in self.filtrations:
            f._teardown()

        # close the writer and reader
        self._writer.close()
        self._reader.close()

    def filter(self):
        """
        Filter the VCF.
        """
        self._logger.info('Start filtering')

        # setup filtrations
        self._setup()

        # get progress bar
        with self.get_pbar(desc=f"{self.__class__.__name__}>Processing sites") as pbar:

            # iterate over the sites
            for i, variant in enumerate(self._reader):

                if self.is_filtered(variant):
                    # write the variant
                    self._writer.write_record(variant)

                pbar.update()

                # explicitly stopping after ``n`` sites fixes a bug with cyvcf2:
                # 'error parsing variant with `htslib::bcf_read` error-code: 0 and ret: -2'
                if i + 1 == self.n_sites or i + 1 == self.max_sites:
                    break

        # teardown filtrations
        self._teardown()

        self._logger.info(f'Filtered out {self.n_filtered} of {self.n_sites} sites in total.')



"""
Utilities for parsing polyDFE files.
"""

__author__ = "Janek Sendrowski"
__contact__ = "sendrowski.janek@gmail.com"
__date__ = "2023-02-26"

import logging
import os
from typing import TextIO, List

import pandas as pd

from .spectrum import Spectrum

# get logger
logger = logging.getLogger('fastdfe').getChild('polydfe')

#: model mapping
models = dict(
    DisplacedGammaParametrization='A',
    GammaDiscreteParametrization='B',
    GammaExpParametrization='C',
    DiscreteParametrization='D'
)

# leading column names in init file
columns = [
    'eps',
    'eps_deprecated',
    'lambda',
    'theta_bar',
    'a'
]


class ParamMode:
    """
    Parameter modes in polyDFE input files.
    """
    independent = 0
    fixed = 1
    shared = 2


def create_sfs_config(
        file: str,
        sfs_neut: Spectrum,
        sfs_sel: Spectrum
):
    """
    Create a sfs config file for polyDFE.

    :param sfs_sel: Selected sfs
    :param sfs_neut: Neutral sfs
    :param file: Path to config file
    """

    def write_line(f: TextIO, sfs: Spectrum, sep: str = '\t'):
        """
        Concatenate given array and write to stream.

        :param sfs: SFS
        :param f: File stream
        :param sep: separator
        """
        # SFS and number of mutational target sites
        # we ignore divergence data here
        data = list(sfs.polymorphic) + [sfs.n_sites]

        f.write(sep.join(map(str, data)) + os.linesep)

    # number of spectra per time
    m_neut = 1
    m_sel = 1
    with open(file, 'w') as fh:
        fh.write(' '.join(map(str, [m_neut, m_sel, sfs_sel.n])) + os.linesep)

        write_line(fh, sfs_neut)
        write_line(fh, sfs_sel)


def parse_init_file(
        dfe_params: list,
        file: str,
        id: int = 1,
        ignore: list = ['eps_deprecated', 'r']
) -> (dict, dict):
    """
    Parse polyDFE init file.
    This will define the initial parameters and
    which ones will be held fixed during the optimization.
    We do not support fixing some of the polyDFE's parameters
    like theta and the nuisance parameters. Some other parameters
    are not supported altogether.

    :param dfe_params: List of dfe parameters
    :param file: Path to init file
    :param id: Init id
    :param ignore: List of parameters to ignore
    :return Dictionary of fixed parameters and dictionary of initial values
    """
    logger.info('Parsing init file.')

    # columns except for nuisance parameters
    cols = columns + list(dfe_params)

    # load init file into dataframe
    df = pd.read_csv(file, delimiter=r"\s+", comment='#', header=None, index_col=0)

    # check if id is present
    if id not in df.index:
        raise Exception(f'Id {id} not found in init file. Possible ids: {str(df.index.to_list())}.')

    # extract row with specified id
    data = df.loc[id]

    # Retrieve parameters and their modes.
    # Note that the parameter order is important
    # when recreating the init file.
    params, params_mode = {}, {}
    for i, col in enumerate(cols):
        params[col] = data[2 * i + 2]
        params_mode[col] = int(data[2 * i + 1])

    # assign nuisance parameters
    params['r'] = list(data[2 * len(cols) + 1:])
    params_mode['r'] = int(data[2 * len(cols) + 1])

    # determine fixed parameters and initial conditions
    fixed_params = {}
    x0 = {}
    for param in params_mode.keys():
        if param not in ignore:
            # here we don't distinguish between shared and independent parameters
            if params_mode[param] == ParamMode.fixed:
                fixed_params[param] = params[param]

            x0[param] = params[param]

    # issue notice
    logger.info(f"Found initial params: {x0} and fixed params: {fixed_params} when parsing init file.")

    return fixed_params, x0


def create_init_file(
        file: str,
        fixed_params: List,
        x0: dict,
        dfe_params: List,
        n: int,
        id: int = 1
):
    """
    Create an init file for polyDFE.

    :param id: Init id
    :param n: SFS sample size
    :param dfe_params: List of dfe parameters
    :param x0: Dictionary of initial values
    :param fixed_params: List of fixed parameters
    :param file: File path to write init file to
    """
    cols = columns + list(dfe_params) + ['r']

    # copy dict to avoid modification
    x0 = x0.copy()

    if 'r' in x0:
        # assign existing nuisance parameters
        x0['r'] = ' '.join([str(r) for r in x0['r']])
    else:
        # make up initial nuisance parameters
        x0['r'] = ' '.join(['1'] * (n - 1))

    # create list of initial parameters
    params = []
    for col in cols:
        if col in x0:
            params += [x0[col]]
        else:
            params += [0]

    # create list of parameter modes
    params_mode = []
    for col in cols:
        if col in fixed_params:
            params_mode += [ParamMode.fixed]
        else:
            params_mode += [ParamMode.independent]

    # concatenate parameters and their modes
    params_with_mode = [str(a) + ' ' + str(b) for a, b in zip(list(params_mode), params)]

    # prepend an id value and join
    line = '\t'.join([str(id)] + params_with_mode)

    # write line to file
    with open(file, 'w') as fh:
        fh.write(line)



"""
MLE utilities.
"""

__author__ = "Janek Sendrowski"
__contact__ = "sendrowski.janek@gmail.com"
__date__ = "2023-02-26"

import numpy as np
from scipy.special import factorial
from scipy.stats import poisson


class Likelihood:
    """
    Utilities for computing Poisson likelihoods.
    """

    #: Epsilon for numerical stability
    eps = 1e-50

    @staticmethod
    def add_epsilon(x: np.ndarray) -> np.ndarray:
        """
        Add epsilon to zero counts.

        :param x: Array to add epsilon to
        :return: Array with epsilon added to zero counts
        """
        # convert to float
        x = x.astype(float)

        # replace 0s with epsilon to avoid log(0)
        x[x == 0] = Likelihood.eps

        return x

    @staticmethod
    def poisson(mu: np.ndarray, k: np.ndarray) -> np.ndarray:
        """
        Compute Poisson(mu, k).

        :param mu: Mean of Poisson distribution
        :param k: Number of events
        :return: Poisson(mu, k)
        """
        # add epsilon to zero counts
        mu = Likelihood.add_epsilon(mu)

        return poisson.pmf(k, mu)

    @staticmethod
    def log_poisson(mu: np.ndarray, k: np.ndarray) -> np.ndarray:
        """
        Compute log(Poisson(mu, k)).

        :param mu: Mean of Poisson distribution
        :param k: Number of events
        :return: log(Poisson(mu, k))
        """
        # add epsilon to zero counts
        mu = Likelihood.add_epsilon(mu)

        return k * np.log(mu) - mu - Likelihood.log_factorial(k)

    @staticmethod
    def log_factorial_stirling(n: np.ndarray | float) -> np.ndarray | float:
        """
        Use Stirling's approximation for values larger than n_threshold.
        https://en.wikipedia.org/wiki/Stirling%27s_approximation

        :param n: n
        :return: log(n!)
        """
        # np.log(np.sqrt(2 * np.pi * n) * (n / np.e) ** n * (1 + 1 / (12 * n)))
        return 0.5 * np.log(2 * np.pi * n) + n * np.log(n / np.e) + np.log(1 + 1 / (12 * n))

    @staticmethod
    def log_factorial(n: np.ndarray, n_threshold: int = 100) -> np.ndarray:
        """
        Compute log(n!).

        :param n: n
        :param n_threshold: Threshold for using Stirling's approximation
        :return: log(n!)
        """
        x = np.zeros_like(n, dtype=np.float64)

        low = n <= n_threshold

        # compute for low values of n
        x[low] = np.log(factorial(n[low]))

        # use Stirling's approximation for large value of n
        x[~low] = Likelihood.log_factorial_stirling(n[~low])

        return x



"""
Optimization module.
"""

__author__ = "Janek Sendrowski"
__contact__ = "sendrowski.janek@gmail.com"
__date__ = "2023-02-26"

import copy
import logging
import math
from dataclasses import dataclass
from typing import Callable, List, Dict, Literal, Tuple, Optional, Sequence

import multiprocess as mp
import numpy as np
from numpy.linalg import norm
from numpy.random import Generator
from scipy.stats import loguniform, uniform
from tqdm import tqdm

from .likelihood import Likelihood
from .settings import Settings

# get logger
logger = logging.getLogger('fastdfe').getChild('Optimization')


def parallelize(
        func: Callable,
        data: Sequence,
        parallelize: bool = True,
        pbar: bool = None,
        desc: str = None,
        dtype: type = object
) -> np.ndarray:
    """
    Parallelize given function or execute sequentially.

    :param parallelize: Whether to parallelize
    :param data: Data to iterate over
    :param func: Function to apply to each element of data
    :param pbar: Whether to show a progress bar
    :param desc: Description for progress bar
    :param dtype: Data type of the returned array
    :return: List of results
    """
    n = len(data)

    if parallelize and n > 1:
        # parallelize
        iterator = mp.Pool().imap(func, data)
    else:
        # sequentialize
        iterator = map(func, data)

    # whether to show a progress bar
    if pbar is True or (pbar is None and n > 1):
        iterator = tqdm(iterator, total=n, disable=Settings.disable_pbar, desc=desc)

    return np.array(list(iterator), dtype=dtype)


def flatten_dict(d: dict, separator='.', prefix=''):
    """
    Flatten dictionary.

    :param d: The nested dictionary
    :param separator: The separator character to use in the flattened dictionary keys
    :param prefix: The prefix to use in the flattened dictionary keys
    :return: The flattened dictionary
    """
    res = {}

    for key, value in d.items():
        if isinstance(value, dict):
            # recursive call
            res.update(flatten_dict(value, separator, prefix + key + separator))
        else:
            res[prefix + key] = value

    return res


def unflatten_dict(d: dict, separator='.'):
    """
    Unflatten dictionary.

    :param d: The flattened dictionary
    :param separator: The separator character used in the flattened dictionary keys
    :return: The original nested dictionary
    """
    res = {}

    for key, value in d.items():
        subkeys = key.split(separator)
        subdict = res

        # recursively create nested dictionaries for each subkey
        for subkey in subkeys[:-1]:
            subdict = subdict.setdefault(subkey, {})

        # assign value to the final subkey
        subdict[subkeys[-1]] = value if not isinstance(value, dict) else unflatten_dict(value, separator)

    return res


def unpack_params(x: np.ndarray, original: Dict[str, dict | tuple | float]) -> Dict[str, dict | tuple | float]:
    """
    Unpack params from numpy array. This is the inverse of pack_params and is used
    as scipy.optimize.minimize only accepts numpy arrays as parameters.

    :param x: Numpy array
    :param original: Original dictionary
    :return: Unpacked dictionary
    """
    keys = flatten_dict(original).keys()

    return unflatten_dict(dict(zip(keys, x)))


def pack_params(params: Dict[str, dict | tuple | float]) -> np.ndarray:
    """
    Pack params into numpy array. This is used as scipy.optimize.minimize only accepts
    numpy arrays as parameters. This is the inverse of unpack_params.

    :param params: Dictionary to pack
    :return: numpy array
    """
    flattened = flatten_dict(params)

    return np.array(list(flattened.values()))


def filter_dict(d, keys):
    """
    Recursively filter a dictionary by a list of given keys at the deepest level.

    :param d: The dictionary to filter
    :param keys: The list of keys to keep in the filtered dictionary
    :return: The filtered dictionary
    """
    filtered = {}

    for key, value in d.items():
        if isinstance(value, dict):
            filtered_sub = filter_dict(value, keys)
            if filtered_sub:
                filtered[key] = filtered_sub
        else:
            if key in keys:
                filtered[key] = value

    return filtered


def pack_shared(
        params: Dict[str, Dict[str, float]],
        shared: List['SharedParams'],
        shared_values: Dict[str, float]
) -> Dict[str, Dict[str, float]]:
    """
    Pack shared parameters. Here we extract shared parameters from
    type-specific keys and instead create entries with joint-types
    holding the shared parameters.
    Note that we only delete parameters from marginal types but in some
    cases we would have to delete them from compound types as well if
    a more general shared parameter were to be specified later on.
    We instead rely on the user not to do this.

    :param params: Dictionary of parameters indexed by type
    :param shared: List of shared parameters
    :param shared_values: Dictionary of parameters indexed by joint-type and parameter name, i.e. 'type1:type2.p1'
    :return: Packed dictionary
    """
    packed = copy.deepcopy(params)

    # iterator through shared parameter
    for s in shared:

        # remove shared keys
        for t in s.types:
            for p in s.params:
                if p in packed[t]:
                    packed[t].pop(p)

        # create joint type string
        type_string = ':'.join(s.types)

        # create shared type with appropriate parameters
        packed = merge_dicts(packed, {type_string: dict((p, shared_values[type_string + '.' + p]) for p in s.params)})

    return packed


def unpack_shared(params: dict) -> dict:
    """
    Unpack shared parameters. Here we extract shared parameters from joint-type keys
    and add them tp the targeted types.

    :param params: Dictionary of parameters
    :return: Unpacked dictionary
    """
    unpacked = {}

    for t, v in params.items():
        if ':' in t:
            unpacked = merge_dicts(unpacked, dict((s, v) for s in t.split(':')))
        else:
            unpacked[t] = v

    return unpacked


def expand_shared(params: List['SharedParams'], types: List[str], names: List[str]) -> List['SharedParams']:
    """
    Expand 'all' type for shared parameters.

    :param params: List of shared parameters
    :param types: List of types
    :param names: List of parameter names
    :return: Expanded list of shared parameters
    """
    expanded = []

    for x in params:
        # expand 'all' type
        types = types if x.types == 'all' else x.types
        params = names if x.params == 'all' else x.params

        # noinspection PyTypeChecker
        expanded.append(SharedParams(types=types, params=params))

    return expanded


def expand_fixed(
        fixed_params: Dict[str, Dict[str, float]],
        types: List[str]
) -> Dict[str, Dict[str, float]]:
    """
    Expand 'all' type for shared parameters.

    :param fixed_params: Dictionary of fixed parameters indexed by type and parameter
    :param types: List of types
    :return: Expanded dictionary of fixed parameters
    """
    expanded = {}

    # loop through fixed parameters
    for key_type, params in fixed_params.items():
        # expand 'all' type
        key_types = types if key_type == 'all' else [key_type]

        # loop through types
        for t in key_types:
            if t not in expanded:
                expanded[t] = {}

            for param, value in params.items():
                expanded[t][param] = value

    return expanded


def collapse_fixed(
        expanded_params: Dict[str, Dict[str, float]],
        types: List[str]
) -> Dict[str, Dict[str, float]]:
    """
    Collapse expanded fixed parameters to 'all' type if all types have the same fixed parameter.

    :param expanded_params: Expanded dictionary of fixed parameters
    :param types: List of types
    :return: Collapsed dictionary of fixed parameters
    """
    all_params = {param: [] for params in expanded_params.values() for param in params}

    # Collect parameter values for all types
    for params in expanded_params.values():
        for param, value in params.items():
            all_params[param].append(value)

    # Calculate the mean and check if parameters can be collapsed
    collapsed = {}
    for param, values in all_params.items():
        if len(values) == len(types):
            collapsed[param] = np.mean(values)

    return {'all': collapsed} if len(collapsed) > 0 else expanded_params


def merge_dicts(dict1: dict, dict2: dict) -> dict:
    """
    Merge two dictionaries recursively.

    :param dict1: First dictionary
    :param dict2: Second dictionary
    :return: Merged dictionary
    """
    # make a copy of the first dictionary
    result = dict(dict1)

    # loop through the items in the second dictionary
    for key, value in dict2.items():

        # Check if the key already exists in the result dictionary and both the
        # value in the result and dict2 dictionaries are dictionaries.
        if key in result and isinstance(result[key], dict) and isinstance(value, dict):

            # recursively merge the two dictionaries
            result[key] = merge_dicts(result[key], value)
        else:
            # simply assign the value from dict2 to the result dictionary
            result[key] = value

    return result


def correct_values(
        params: Dict[str, float],
        bounds: Dict[str, Tuple[float, float]],
        scales: Dict[str, Literal['lin', 'log', 'symlog']],
        warn: bool = False,
        threshold: float = 1e-6
) -> Dict[str, float]:
    """
    Correct initial values so that they are within the specified bounds.

    :param bounds: Dictionary of bounds
    :param params: Flattened dictionary of parameters
    :param scales: Dictionary of scales
    :param warn: Whether to warn if values are corrected
    :param threshold: Threshold for the error to trigger a warning
    :return: Corrected dictionary
    """
    # create a copy of params
    corrected = params.copy()

    for key, value in params.items():
        # get base name
        name = key.split('.')[-1]

        # get real bounds
        bound = get_real_bounds(bounds[name], scale=scales[name])

        # correct value if outside bounds
        if value < bound[0]:
            corrected[key] = bound[0]
        elif value > bound[1]:
            corrected[key] = bound[1]

    # differences between the original and corrected dictionaries
    differences = {key: (params[key], corrected[key]) for key in params if params[key] != corrected[key]}

    # warn if there are differences that exceed the threshold
    exceeded_threshold = {}
    for key, (old_val, new_val) in differences.items():

        # calculate relative error
        err = np.abs(new_val - old_val)

        # add if it exceeds the relative error
        if err > threshold:
            exceeded_threshold[key] = f"{old_val} -> {new_val}"

    if exceeded_threshold and warn:
        logger.warning(f'Given initial values outside bounds. Adjusting {exceeded_threshold}.')

    return corrected


def get_real_bounds(bounds: Tuple[float, float], scale: Literal['lin', 'log', 'symlog']) -> Tuple[float, float]:
    """
    Get real bounds from the given bounds.

    :param bounds: Bounds of the parameter
    :param scale: Scale of the parameter
    :return:
    """
    if scale == 'symlog':
        return -bounds[1], bounds[1]

    return bounds


def evaluate_counts(get_counts: dict, params: dict):
    """
    Evaluate counts using the given parameters.
    Here we assign the parameters to the appropriate types
    obtaining the counts for each type.

    :param get_counts: Dictionary of functions to evaluate counts for each type
    :param params: Dictionary of parameters
    :return: Dictionary of counts
    """
    counts = {}

    # unpack shared parameters
    unpacked = unpack_shared(params)

    # evaluate counts for each type
    for key in get_counts.keys():
        counts[key] = get_counts[key](unpacked[key])

    return counts


def to_symlog(x: float, linthresh: float = 1e-5) -> float:
    """
    Convert a value to the symlog scale.

    :param x: The input value on the original scale.
    :param linthresh: The positive value that determines the range within which the
        symlog scale is linear. Must be greater than 0.
    :return: The value on the symlog scale.
    """
    sign = np.sign(x)
    abs_x = np.abs(x)
    log_x = np.log10(abs_x + linthresh) - np.log10(linthresh)

    return sign * (abs_x / linthresh if abs_x <= linthresh else log_x)


def from_symlog(y: float, linthresh: float = 1e-5) -> float:
    """
    Convert a value from the symlog scale back to the original scale.

    :param y: The input value on the symlog scale.
    :param linthresh: The positive value that determines the range within which the
        symlog scale is linear. Must be greater than 0.
    :return: The value on the original scale.
    """
    sign = np.sign(y)
    abs_y = np.abs(y)
    exp_y = np.power(10, abs_y + np.log10(linthresh)) - linthresh

    return sign * (abs_y * linthresh if abs_y <= 1 else exp_y)


def scale_bound(bounds: Tuple[float, float], scale: Literal['lin', 'log', 'symlog']):
    """
    Convert a bound to the specified scale. For symlog scale we assume the symmetric bounds,
    so that the upper bound denotes the boundaries and the lower bound the linear threshold.

    :param bounds: The bound to convert
    :param scale: The scale to convert to
    :return: The converted bound
    :raises ValueError: if the scale is unknown
    """
    if scale == 'lin':
        return bounds

    if scale == 'log':
        if bounds[1] < 0:
            return -np.log10(-bounds[0]), -np.log10(-bounds[1])

        if bounds[0] > 0:
            return np.log10(bounds[0]), np.log10(bounds[1])

        raise ValueError('Bounds must not span zero for log scale.')

    if scale == 'symlog':

        if bounds[0] <= 0 or bounds[1] <= 0:
            raise ValueError('Both bounds must be positive for symlog scale.')

        return to_symlog(-bounds[1], linthresh=bounds[0]), to_symlog(bounds[1], linthresh=bounds[0])

    raise ValueError(f'Unknown scale {scale}.')


def unscale_bound(
        scaled_bounds: Tuple[float, float],
        scale: Literal['lin', 'log', 'symlog'],
        linthresh: float = 1e-5
) -> Tuple[float, float]:
    """
    Convert a bound from the specified scale back to the original scale. For symlog scale,
    we assume symmetric bounds, so that the upper bound denotes the boundaries and
    the lower bound the linear threshold, i.e. ``bounds = (-bounds[1], bounds[1])`` and ``linthresh = bounds[0]``.
    Note that we cannot reliably recover negative bounds that were log scaled.

    :param linthresh:
    :param scaled_bounds: The bound to convert
    :param scale: The scale to convert from
    :return: The converted bound
    :raises ValueError: if the scale is unknown
    """
    if scale == 'lin':
        return scaled_bounds

    if scale == 'log':
        return np.power(10, scaled_bounds[0]), np.power(10, scaled_bounds[1])

    if scale == 'symlog':
        upper_bound = from_symlog(scaled_bounds[1], linthresh=linthresh)

        return linthresh, upper_bound

    raise ValueError(f'Unknown scale {scale}.')


def scale_bounds(
        bounds: Dict[str, Tuple[float, float]],
        scales: Dict[str, Literal['lin', 'log', 'symlog']]
) -> Dict[str, Tuple[float, float]]:
    """
    Convert bounds to the specified scale. For symlog scale we assume the symmetric bounds,
    so that the upper bound denotes the boundaries and the lower bound the linear threshold.

    :param bounds: Flattened dictionary of bounds to convert index by type and parameter
    :param scales: Dictionary of scales indexed by parameter
    :return: The converted bounds
    :raises ValueError: if the scale is unknown
    """
    scaled_bounds = {}

    for key, value in bounds.items():
        scaled_bounds[key] = scale_bound(value, scale=scales[get_basename(key)])

    return scaled_bounds


def scale_value(value: float, bounds: Tuple[float, float], scale: Literal['lin', 'log', 'symlog']) -> float:
    """
    Convert a value to the specified scale. For symlog scale, the untransformed bounds are needed,
    so that the upper bound denotes the boundaries and the lower bound the linear threshold.

    :param value: The value to convert.
    :param bounds: The untransformed bounds for the symlog scale.
    :param scale: The scale to convert to.
    :return: The converted value.
    :raises ValueError: if the scale is unknown.
    """
    if scale == 'lin':
        return value

    if scale == 'log':

        if value < 0:
            return -np.log10(-value)

        return np.log10(value)

    if scale == 'symlog':
        return to_symlog(value, linthresh=bounds[0])

    raise ValueError(f'Unknown scale {scale}.')


def unscale_value(scaled_value: float, bounds: Tuple[float, float], scale: Literal['lin', 'log', 'symlog']) -> float:
    """
    Convert a value from the specified scale back to the original scale. For symlog scale,
    the untransformed bounds are needed, so that the upper bound denotes the boundaries
    and the lower bound the linear threshold.

    :param scaled_value: The value to convert.
    :param bounds: The untransformed bounds for the symlog scale.
    :param scale: The scale to convert from.
    :return: The converted value.
    :raises ValueError: if the scale is unknown.
    """
    if scale == 'lin':
        return scaled_value

    if scale == 'log':

        if bounds[1] < 0:
            return -np.power(10, -scaled_value)

        return np.power(10, scaled_value)

    if scale == 'symlog':
        return from_symlog(scaled_value, linthresh=bounds[0])

    raise ValueError(f'Unknown scale {scale}.')


def scale_values(
        params: Dict[str, Dict[str, float]],
        bounds: Dict[str, Tuple[float, float]],
        scales: Dict[str, Literal['lin', 'log', 'symlog']]
) -> Dict[str, Dict[str, float]]:
    """
    Scale values according to the given scales.

    :param params: Nested dictionary of parameters indexed by type and parameter
    :param scales: Dictionary of scales indexed by parameter name
    :param bounds: Dictionary of bounds indexed by parameter name
    :return: Nested dictionary of scaled parameters indexed by type and parameter
    """
    scaled = {}

    for key, value in flatten_dict(params).items():
        # scale value
        scaled[key] = scale_value(value, bounds[get_basename(key)], scales[get_basename(key)])

    return unflatten_dict(scaled)


def unscale_values(
        params: Dict[str, Dict[str, float]],
        bounds: Dict[str, Tuple[float, float]],
        scales: Dict[str, Literal['lin', 'log', 'symlog']]
) -> Dict[str, Dict[str, float]]:
    """
    Unscale values according to the given scales.

    :param params: Nested dictionary of parameters indexed by type and parameter
    :param scales: Dictionary of scales indexed by parameter name
    :param bounds: Dictionary of scales indexed by parameter name
    :return: Nested dictionary of unscaled parameters indexed by type and parameter
    """
    unscaled = {}

    for key, value in flatten_dict(params).items():
        # unscale value
        unscaled[key] = unscale_value(value, bounds[get_basename(key)], scales[get_basename(key)])

    return unflatten_dict(unscaled)


def get_basename(name: str) -> str:
    """
    Get the basename of parameter string, i.e. type.param -> param.

    :param name: The string to get the basename from.
    :return: The basename.
    """
    return name.split('.')[-1]


def check_bounds(
        bounds: Dict[str, Tuple[float, float]],
        params: Dict[str, float],
        fixed_params: Dict[str, float] = {},
        percentile: float = 1,
        scale: Literal['lin', 'log'] = 'lin'
) -> Tuple[Dict[str, Tuple[float, float, float]], Dict[str, Tuple[float, float, float]]]:
    """
    Issue warnings if the passed parameters are close to the specified bounds.

    :param bounds: The bounds to check against.
    :param params: The parameters to check.
    :param fixed_params: The fixed parameters.
    :param percentile: The percentile threshold to consider a parameter close to the bounds.
    :param scale: Scale type: 'lin' for linear and 'log' for logarithmic.
    :return: Tuple of dictionaries of parameters close to the lower and upper bounds, i.e. (lower, value, upper).
    """
    near_lower = {}
    near_upper = {}

    def transform(value: float, to_scale: Literal['lin', 'log']) -> float:
        """
        Transform a value to the specified scale.

        :param value: The value to transform.
        :param to_scale: The scale to transform to.
        :return: The transformed value.
        """
        if to_scale == 'log':
            return math.log(value) if value > 0 else -float('inf')

        return value

    for key, value in params.items():
        # get base name
        name = key.split('.')[-1]

        # get bounds
        lower, upper = bounds[name]

        # transform values
        _lower = transform(lower, scale)
        _upper = transform(upper, scale)
        _value = transform(value, scale)

        if key not in fixed_params:
            if _lower is not None and (_value - _lower) / (_upper - _lower) <= percentile / 100:
                near_lower[key] = (lower, value, upper)

            if _upper is not None and (_upper - _value) / (_upper - _lower) <= percentile / 100:
                near_upper[key] = (lower, value, upper)

    return near_lower, near_upper


@dataclass
class SharedParams:
    """
    Class specifying the sharing of params among types.
    ``all`` means all available types or params.

    Example usage:

    ::

        import fastdfe as fd

        # neutral SFS for two types
        sfs_neut = fd.Spectra(dict(
            pendula=[177130, 997, 441, 228, 156, 117, 114, 83, 105, 109, 652],
            pubescens=[172528, 3612, 1359, 790, 584, 427, 325, 234, 166, 76, 31]
        ))

        # selected SFS for two types
        sfs_sel = fd.Spectra(dict(
            pendula=[797939, 1329, 499, 265, 162, 104, 117, 90, 94, 119, 794],
            pubescens=[791106, 5326, 1741, 1005, 756, 546, 416, 294, 177, 104, 41]
        ))

        # create inference object
        inf = fd.JointInference(
            sfs_neut=sfs_neut,
            sfs_sel=sfs_sel,
            shared_params=[fd.SharedParams(types=["pendula", "pubescens"], params=["eps", "S_d"])],
            do_bootstrap=True
        )

        # run inference
        inf.run()

    """
    #: The params to share
    params: List[str] | Literal['all'] = 'all'

    #: The types to share
    types: List[str] | Literal['all'] = 'all'


@dataclass
class Covariate:
    """
    Class defining a covariate which induces a relationship
    with one or many parameters. The relationship is defined
    by a callback function which modifies the parameters. The
    default callback introduces a linear relationship.

    Below an example of introducing linear covariates for ``S_d``, the for mean strength of deleterious
    mutations (cf. :class:`~fastdfe.parametrization.GammaExpParametrization`). Each
    of the three types is associated with one covariate. This we pass to
    :class:`~fastdfe.joint_inference.JointInference` together with the stratified spectra:

    ::

        import fastdfe as fd

        cov = fd.Covariate(
            param='S_d',
            values=dict(type1=5, type2=3, type3=1)
        )

    """

    #: The parameter to modify
    param: str

    #: The values of the covariate for each type
    values: Dict[str, float]

    #: The callback function to modify the parameters
    callback: Optional[Callable] = None

    #: The bounds of the covariate parameter to be estimated
    bounds: tuple = (1e-4, 1e4)

    #: The initial value of the covariate
    x0: float = 0

    #: The scale of the bounds. See :func:`scale_value` for details
    bounds_scale: Literal['lin', 'log', 'symlog'] = 'symlog'

    def __post_init__(self):
        """
        Cast bounds to tuple and check if an inverse_callback is provided
        when a custom callback is specified.
        """
        self.bounds = tuple(self.bounds)

    def apply(self, covariate: float, type: str, params: Dict[str, float]) -> Dict[str, float]:
        """
        Apply the custom or default callback to modify the given parameters.

        :param covariate: The value of the covariate.
        :param type: The type of the relationship.
        :param params: The input parameters.
        :return: Modified parameters.
        """
        # Use custom callback if given else default callback
        callback = self.apply_default if self.callback is None else self.callback

        return callback(covariate=covariate, type=type, params=params)

    def apply_default(self, covariate: float, type: str, params: Dict[str, float]) -> Dict[str, float]:
        """
        Modify the given parameters introducing a linear relationship
        with the given covariate.

        :param covariate: The value of the covariate.
        :param type: The type of the relationship.
        :param params: The input parameters.
        :return: Modified parameters.
        """
        # create a copy of input parameters
        modified = params.copy()

        # introduce linear relationship
        if self.param in params:
            modified[self.param] += covariate * self.values[type]

        return modified

    @staticmethod
    def _apply(covariates: Dict[str, 'Covariate'], params: dict, type: str) -> dict:
        """
        Apply given covariates to given parameters.

        :param covariates: Dictionary of covariates to add
        :param params: Dict of parameters
        :param type: SFS type
        :return: Dict of parameters with covariates added
        """
        for k, cov in covariates.items():
            params = cov.apply(
                covariate=params[k],
                type=type,
                params=params
            )

        return params


class Optimization:
    """
    Class for optimizing the DFE.
    """

    #: Optimization method to use. Use class property for downward compatibility
    method_mle = 'L-BFGS-B'

    def __init__(
            self,
            bounds: Dict[str, Tuple[float, float]],
            param_names: List[str],
            loss_type: Literal['likelihood', 'L2'] = 'likelihood',
            opts_mle: dict = {},
            method_mle: str = 'L-BFGS-B',
            parallelize: bool = True,
            fixed_params: Dict[str, Dict[str, float]] = {},
            scales: Dict[str, Literal['lin', 'log', 'symlog']] = {},
            seed: int = None
    ):
        """
        Create object.

        :param parallelize: Whether to parallelize the optimization
        :param bounds: Dictionary of bounds
        :param opts_mle: Dictionary of options for the optimizer
        :param method_mle: Optimization method to use
        :param loss_type: Type of loss function to use
        :param fixed_params: Dictionary of fixed parameters
        :param scales: Dictionary of scales
        :param param_names: List of parameter names
        """
        #: Parameter bounds
        self.bounds = bounds

        #: Parameter scales to use
        self.scales = scales

        #: additional options for the optimizer
        self.opts_mle = opts_mle

        #: Optimization method to use
        self.method_mle = method_mle

        #: Type of loss function to use
        self.loss_type = loss_type

        #: Fixed parameters
        self.fixed_params = flatten_dict(fixed_params)

        # check if fixed parameters are within the specified bounds
        if correct_values(self.fixed_params, self.bounds, warn=False, scales=scales) != self.fixed_params:
            raise ValueError('Fixed parameters are outside the specified bounds. '
                             f'Fixed params: {self.fixed_params}, bounds: {self.bounds}.')

        #: Parameter names
        self.param_names = param_names

        #: Whether to parallelize the optimization
        self.parallelize = parallelize

        #: Initial values
        self.x0: Optional[dict] = None

        #: Number of runs
        self.n_runs: Optional[int] = None

        #: Likelihoods for each run
        self.likelihoods: Optional[np.ndarray] = None

        #: Random generator instance
        self.rng = np.random.default_rng(seed=seed)

    def run(
            self,
            get_counts: Dict[str, Callable],
            x0: Dict[str, Dict[str, float]] = {},
            scales: Dict[str, Literal['lin', 'log', 'symlog']] = {},
            bounds: Dict[str, Tuple[float, float]] = {},
            n_runs: int = 1,
            debug_iterations: bool = True,
            print_info: bool = True,
            opts_mle: dict = None,
            pbar: bool = None,
            desc: str = 'Inferring DFE',
    ) -> ('scipy.optimize.OptimizeResult', dict):
        """
        Perform the optimization procedure.

        :param scales: Scales of the parameters
        :param bounds: Bounds of the parameters
        :param n_runs: Number of independent optimization runs out of which the best one is chosen. The first run
            will use the initial values if specified. Consider increasing this number if the optimization does not
            produce good results.
        :param x0: Dictionary of initial values in the form ``{type: {param: value}}``
        :param get_counts: Dictionary of functions to evaluate counts for each type
        :param debug_iterations: Whether to print debug messages for each iteration
        :param opts_mle: Dictionary of options for the optimizer
        :param print_info: Whether to print information about the bounds
        :param pbar: Whether to show a progress bar
        :param desc: Description for the progress bar
        :return: The optimization result and the likelihoods
        """
        from scipy.optimize import minimize, OptimizeResult

        # number of optimization runs
        self.n_runs = n_runs

        # store the scales of the parameters
        if scales:
            self.scales = scales

        # store the bounds of the parameters
        if bounds:
            self.bounds = bounds

        # store the options for the optimizer
        if opts_mle:
            self.opts_mle = opts_mle

        # filter out unneeded values
        # this also holds the fixed parameters
        self.x0 = filter_dict(x0, self.param_names)

        # flatten initial values
        flattened = flatten_dict(self.x0)

        # determine parameter names of parameters to be optimized
        optimized_param_names = list(set(flattened) - set(self.fixed_params))

        # issue debug messages
        logger.debug(f'Performing optimization on {len(flattened)} parameters: {list(flattened.keys())}.')
        logger.debug(f'Using initial values: {flattened}.')
        logger.debug(f"Optimizing parameters: {optimized_param_names}")

        # issue warning when the number of parameters to be optimized is large
        if len(optimized_param_names) > 10 and print_info:
            logger.warning(f'A large number of parameters is optimized jointly ({len(optimized_param_names)}). '
                           f'Please be aware that this makes it harder to find a good optimum.')

        # correct initial values to be within bounds
        self.x0 = unflatten_dict(correct_values(flattened, self.bounds, warn=True, scales=self.scales))

        # determine parameter bounds
        bounds = self.get_bounds(flatten_dict(self.x0))

        def optimize(x0: Dict[str, Dict[str, float]]) -> OptimizeResult:
            """
            Perform numerical minimization.

            :param x0: Dictionary of initial values in the form ``{type: {param: value}}``
            :return: Optimization result
            """
            logger.debug(f"Initial parameters: {x0}.")

            return minimize(
                fun=self.get_loss_function(
                    get_counts=get_counts,
                    print_debug=debug_iterations
                ),
                x0=pack_params(self.scale_values(x0)),
                method=self.method_mle,
                bounds=pack_params(scale_bounds(bounds, self.scales)),
                options=self.opts_mle
            )

        # initial parameters for the samples
        initial_params = [self.x0] + [self.sample_x0(self.x0) for _ in range(int(self.n_runs) - 1)]

        # parallelize MLE for different initializations
        results = parallelize(optimize, initial_params, self.parallelize, pbar=pbar, desc=desc)

        # list of the best likelihood for each run
        self.likelihoods = -np.array([res.fun for res in results])

        # get result with the lowest likelihood
        result = results[np.argmax(self.likelihoods)]

        # unpack MLE params array into a dictionary
        params_mle = unpack_params(result.x, self.x0)

        # unscale parameters
        params_mle = unscale_values(params_mle, self.bounds, self.scales)

        # check if the MLE reached one of the bounds
        if print_info:
            self.check_bounds(flatten_dict(params_mle))

        return result, params_mle

    def scale_values(self, values: Dict[str, Dict[str, float]]) -> Dict[str, Dict[str, float]]:
        """
        Scale the values of the parameters.

        :param values: Dictionary of initial values in the form ``{type: {param: value}}``
        :return: Dictionary of scaled initial values
        """
        return scale_values(values, self.bounds, self.scales)

    def get_loss_function(
            self,
            get_counts: Dict[str, Callable],
            print_debug: bool = True
    ) -> Callable:
        """
        Get the loss function.

        :param get_counts: Dictionary of functions to evaluate counts for each type
        :param print_debug: Whether to print debug messages
        :return: The loss function
        """

        def loss(x: np.ndarray) -> float:
            """
            The loss function.

            :param x: Parameters
            :return: The loss
            """
            # unpack parameters into dictionary using the keys of self.x0
            params = unpack_params(x, self.x0)

            # unscale parameters
            params = unscale_values(params, self.bounds, self.scales)

            # Model SFS from parameters.
            # Here the order of types does not matter.
            # We only collect the counts for types that are
            # given in get_counts. This makes it possible to
            # avoid specifying type 'all' which is of no use
            # in joint inference.
            counts_dict = evaluate_counts(get_counts, params)

            # flatten and convert to array
            counts = np.array(list(counts_dict.values()))

            # reshape and merge
            counts_modelled, counts_observed = np.stack(counts, axis=1).reshape(2, -1)

            # use independent Poisson likelihoods
            LL = Likelihood.log_poisson(mu=counts_modelled, k=counts_observed)

            # combine likelihoods
            ll = np.sum(LL)

            # compute L2 norm
            L2 = norm(counts_modelled - counts_observed, 2)

            # information on iteration
            iter_info = flatten_dict(params) | dict(likelihood=ll, L2=L2)

            # log likelihood
            if print_debug:
                # check likelihood
                if np.isnan(ll):
                    raise ValueError('Oh boy, likelihood is nan. This is no good...')

                # log variables
                logger.debug(iter_info)

            # return appropriate loss
            return dict(L2=L2, likelihood=-ll)[self.loss_type]

        return loss

    def sample_x0(self, example: dict) -> Dict[str, dict]:
        """
        Sample initial values.

        :param example: An example dictionary for generating the initial values
        :return: A dictionary of initial values
        """
        sample = {}

        for key, value in example.items():
            if isinstance(value, dict):
                sample[key] = self.sample_x0(value)
            else:
                sample[key] = self.sample_value(self.bounds[key], self.scales[key], random_state=self.rng)

        return sample

    @staticmethod
    def sample_value(
            bounds: Tuple[float, float],
            scale: Literal['lin', 'log', 'symlog'],
            random_state: int | Generator = None
    ) -> float:
        """
        Sample a value between given bounds using the given scaling.
        This function works for positive, negative, and mixed bounds.
        Note that when ``scale == 'symlog'``, ``bounds[0]`` defines the linear threshold and
        the actual bounds are ``(-bounds[1], bounds[1])``.

        :param bounds: Tuple of lower and upper bounds
        :param scale: Scaling of the parameter.
        :param random_state: Random state
        :return: Sampled value
        """

        def flip(bounds: Tuple[float, float]) -> Tuple[float, float]:
            """
            Flip the bounds.

            :param bounds: Tuple of lower and upper bounds
            :return: Flipped bounds
            """
            return -bounds[1], -bounds[0]

        def symlog_rvs(lower: float, upper: float, random_state: int | Generator = None) -> float:
            """
            Sample from a symmetric log-uniform distribution.

            :param lower: Lower bound which is the linear threshold
            :param upper: Upper bound so that the actual bounds are (-upper, upper)
            :param random_state: Random state
            :return: Sampled value
            """
            val = loguniform.rvs(lower, upper, random_state=random_state)

            # flip sign with 50% probability
            return val if uniform.rvs() < 0.5 else -val

        # dictionary of scaling functions
        scaling_functions = {
            'lin': uniform.rvs,
            'log': loguniform.rvs,
            'symlog': symlog_rvs
        }

        # raise an error if the scale is not valid
        if scale not in scaling_functions:
            raise ValueError(f"Scale must be one of: {', '.join(scaling_functions.keys())}")

        # raise an error if bounds span 0 and scale is 'log'
        if bounds[0] < 0 < bounds[1] and scale == 'log':
            raise ValueError(f"Log scale not possible for bounds that span 0.")

        # raise an error if bounds are negative and scale is 'symlog'
        if bounds[0] < 0 and scale == 'symlog':
            raise ValueError(f"Symlog scale not possible for negative bounds.")

        # flip bounds if they are negative
        flipped = bounds[0] < 0
        if flipped:
            bounds = flip(bounds)

        # sample a value using the appropriate scaling function
        sample = scaling_functions[scale](bounds[0], bounds[1] - bounds[0], random_state=random_state)

        # return the sampled value, flipping back if necessary
        return -sample if flipped else sample

    def check_bounds(self, params: Dict[str, float], percentile: float = 1) -> None:
        """
        Check if the given parameters are within the bounds.

        :param params: Parameters
        :param percentile: Percentile of the bounds to check
        :return: Whether the parameters are within the bounds
        """
        # we scale the bounds to obtain more sensible warnings
        bounds = scale_bounds(self.bounds, self.scales)
        params_scaled = flatten_dict(self.scale_values(unflatten_dict(params)))

        # get parameters close to the bounds
        near_lower, near_upper = check_bounds(
            params=params_scaled,
            bounds=bounds,
            fixed_params=self.fixed_params,
            percentile=percentile,
            scale='lin'
        )

        if len(near_lower | near_upper) > 0:

            def get_values(keys: List[str]) -> Dict[str, Tuple[str, str, str]]:
                """
                Unscale the parameters.

                :param keys: List of parameter names
                :return: Unscaled parameters
                """
                unscaled = {}

                for key in keys:
                    unscaled[key] = (
                        "{:g}".format(self.bounds[get_basename(key)][0]),
                        "{:.8g}".format(params[key]),
                        "{:g}".format(self.bounds[get_basename(key)][1])
                    )

                return unscaled

            # string representation of parameters
            near_lower_unscaled = str(get_values(list(near_lower.keys()))).replace('\'', '')
            near_upper_unscaled = str(get_values(list(near_upper.keys()))).replace('\'', '')

            # issue warning
            logger.warning(
                f'The MLE estimate is close to the upper bound '
                f'for {near_upper_unscaled} and lower bound '
                f'for {near_lower_unscaled} [(lower, value, upper)], but '
                f'this might be nothing to worry about.'
            )

    def get_bounds(self, x0: Dict[str, float]) -> Dict[str, Tuple[float, float]]:
        """
        Get a nested dictionary of bounds the same structure as the given initial values.

        :param x0: Flattened dictionary of initial values
        :return: A dictionary of initial values
        """
        bounds = {}

        for key, value in x0.items():

            # check if the parameter is fixed
            if key in self.fixed_params:
                bounds[key] = (self.fixed_params[key], self.fixed_params[key])
            else:
                bounds[key] = self.bounds[key.split('.')[-1]]

        return bounds

    def set_fixed_params(self, fixed_params: Dict[str, Dict[str, float]]):
        """
        Set fixed parameters. We flatten the dictionary to make it easier to work with.

        :param fixed_params: Dictionary of fixed parameters
        """
        self.fixed_params = flatten_dict(fixed_params)
